# window.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Import libs
import os
from gi.repository import Gtk, Gdk, Gio, GdkPixbuf, GLib

# Import various classes
from .image import DrImage
from .new_image_dialog import DrCustomImageDialog
from .minimap import DrMinimap
from .options_manager import DrOptionsManager
from .message_dialog import DrMessageDialog
from .deco_manager import DrDecoManagerMenubar, \
                          DrDecoManagerHeaderbar, \
                          DrDecoManagerToolbar
from .saving_manager import DrSavingManager
from .printing_manager import DrPrintingManager
from .tools_initializer import DrToolsInitializer

# Import various functions
from .utilities_files import utilities_add_filechooser_filters, \
                             utilities_gfile_is_image

UI_PATH = '/com/github/maoschanz/drawing/ui/'
DEFAULT_TOOL_ID = 'pencil'

################################################################################

@Gtk.Template(resource_path=UI_PATH+'window.ui')
class DrWindow(Gtk.ApplicationWindow):
	__gtype_name__ = 'DrWindow'

	gsettings = Gio.Settings.new('com.github.maoschanz.drawing')

	# Window empty widgets
	tools_flowbox = Gtk.Template.Child()
	toolbar_box = Gtk.Template.Child()
	info_bar = Gtk.Template.Child()
	info_label = Gtk.Template.Child()
	info_action = Gtk.Template.Child()
	notebook = Gtk.Template.Child()
	bottom_panes_box = Gtk.Template.Child()
	unfullscreen_btn = Gtk.Template.Child()
	bottom_meta_box = Gtk.Template.Child()
	tools_scrollable_box = Gtk.Template.Child()
	tools_nonscrollable_box = Gtk.Template.Child()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.app = kwargs['application']

		self.pointer_to_current_page = None # this ridiculous hack allows to
		                   # manage several tabs in a single window despite the
		                                      # notebook widget being pure shit
		self.active_tool_id = None
		self._is_tools_initialisation_finished = False
		self.devel_mode = False
		self.should_track_framerate = False

		if self.gsettings.get_boolean('maximized'):
			self.maximize()

		self._update_theme_variant()
		# self.resize(360, 648)
		# self.resize(720, 288)
		self._set_ui_bars()

	def init_window_content_async(self, content_params):
		"""Initialize the window's content, such as the minimap, the color
		popovers, the tools, their options, and a new image. Depending on the
		parameters, the new image can be imported from the clipboard, loaded
		from a GioFile, or (else) it can be a blank image.

		This method is called asynchronously, which isn't *correct* (not very
		thread-safe or anything) but it allows the window to be shown quicker.
		If it fails, a window is here anyway because this is independent from
		the object constructor."""

		self.reveal_action_report(_("Error starting the application, please" + \
		                                                   " report this bug."))

		gfile = content_params['gfile']
		get_cb = content_params['get_cb']

		self.tools = {}
		self.minimap = DrMinimap(self, None)
		self.options_manager = DrOptionsManager(self)
		self.saving_manager = DrSavingManager(self)
		self.printing_manager = DrPrintingManager(self)

		self.devel_mode = self.gsettings.get_boolean('devel-only')
		self.add_all_win_actions()
		self._init_tools()
		self.connect_signals()

		# The picture is built as late as possible in the init process because
		# reading files is too prone to exceptions that may fuck everything up,
		# and i don't trust that catching them is enough to ensure the process
		# can continue normally.
		try:
			if get_cb:
				self.delayed_build_from_clipboard()
			elif gfile is not None:
				self._build_new_tab(gfile=gfile)
			else:
				self.build_blank_image()
		except Exception as excp:
			self.reveal_message(str(excp))

		self.active_tool().select_flowbox_child()
		self._is_tools_initialisation_finished = True

		self._try_show_release_notes()

		# has to return False to be removed from the mainloop immediately
		return False

	def _try_show_release_notes(self):
		"""Add specific actions to the primary menu if the current version is a
		new one (the corresponding release notes have never been opened or
		dismissed)"""
		last_version = self.gsettings.get_string('last-version')
		current_version = self.app.get_current_version()
		if current_version == last_version:
			return

		# To avoid spamming users too much, only major updates are worthy of
		# that message dialog.
		if last_version.split('.')[0] == current_version.split('.')[0] \
		and last_version.split('.')[1] == current_version.split('.')[1]:
			self.gsettings.set_string('last-version', current_version)
			return

		dialog = DrMessageDialog(self)
		# First use of the app
		if last_version == '0.0.0':
			dialog.add_string(_("It's the first time you use Drawing, " + \
			                         "would you like to read the help manual?"))

		# Major update (assuming i actually follow semantic versionning lol)
		else:
			# Context: %s is the version number of the app
			label = _("It's the first time you use Drawing %s, " + \
			                               "would you like to read what's new?")
			dialog.add_string(label % current_version)
			dialog.add_string(_("These news will always be available in the" + \
			     " help manual: click the 'Help' menu item whenever you want!"))

		no_id = dialog.set_action(_("No"), None, True)
		yes_id = dialog.set_action(_("Yes"), 'suggested-action')
		result = dialog.run()
		dialog.destroy()

		if result == yes_id:
			if last_version == '0.0.0':
				self.app.on_help_index()
			else:
				self.app.on_help_whats_new()
		self.gsettings.set_string('last-version', current_version)

	def _init_tools(self):
		"""Initialize all tools, building the UI for them including the menubar,
		and enable the default tool."""
		disabled_tools = self.gsettings.get_strv('disabled-tools')
		dev = self.gsettings.get_boolean('devel-only')
		self.tools = {}
		self.log_message('window has started, now loading tools')
		self._hide_message()

		tools_initializer = DrToolsInitializer(self)
		self.tools = tools_initializer.load_all_tools(dev, disabled_tools)

		# Side pane items for tools
		self.tools_flowbox.connect('selected-children-changed', \
		                           self._update_active_tool_from_flowbox_signal)
		for tool_id in self.tools:
			self.tools[tool_id].build_flowbox_child(self.tools_flowbox)
		self.on_show_labels_setting_changed()

		# Tools's menubar items if they don't exist yet (they're defined on the
		# application level, so they should only be built the first time)
		if not self.app.has_tools_in_menubar:
			self.build_menubar_tools_menu()

		# Initialisation of which tool is active
		tool_id = self.gsettings.get_string('last-active-tool')
		if tool_id not in self.tools:
			tool_id = DEFAULT_TOOL_ID
		self.active_tool_id = tool_id
		self.former_tool_id = tool_id
		# the end of this process will happen later because it requires an
		# active image, which doesn't exist at this point of the init process.

	def _update_active_tool_from_flowbox_signal(self, *args):
		selected_tool = self._get_newly_selected_tool()
		self.switch_to(selected_tool.id)

	def _get_newly_selected_tool(self):
		for tool_id in self.tools:
			if self.tools[tool_id].is_flowbox_child_selected():
				return self.tools[tool_id]

	def build_menubar_tools_menu(self):
		sections = [None, None, None]
		# the ids of the various kinds of tools don't match the index of their
		# section in the submenu
		sections[2] = self._get_menubar_tools_section(0)
		sections[0] = self._get_menubar_tools_section(1)
		sections[1] = self._get_menubar_tools_section(2)
		for tool_id in self.tools:
			tool = self.tools[tool_id]
			tool.add_item_to_menu(sections[tool.menu_id])
		self.app.has_tools_in_menubar = True

	def _get_menubar_tools_section(self, section_index):
		return self._get_menubar_item([
			[True, 4],
			[False, 1],
			[True, 0],
			[False, section_index]
		])

	############################################################################
	# TABS AND WINDOWS MANAGEMENT ##############################################

	def build_blank_image(self, *args):
		"""Open a new tab with a drawable blank image using the default values
		defined by user's settings."""
		width = self.gsettings.get_int('default-width')
		height = self.gsettings.get_int('default-height')
		rgba = self.gsettings.get_strv('default-rgba')
		self._build_new_tab(width=width, height=height, background_rgba=rgba)
		self.update_picture_title()

	def build_new_custom(self, *args):
		"""Open a new tab with a drawable blank image using the custom values
		defined by user's input."""
		dialog = DrCustomImageDialog(self)
		result = dialog.run()
		if result == Gtk.ResponseType.OK:
			width, height, rgba = dialog.get_values()
			self._build_new_tab(width=width, height=height, background_rgba=rgba)
			self.update_picture_title()
		dialog.destroy()

	def delayed_build_from_clipboard(self, *args):
		"""Calls `async_build_from_clipboard` asynchronously."""
		self._build_new_tab() # temporary image to avoid errors when the window
		# finishes its initialisation.
		GLib.timeout_add(500, self.async_build_from_clipboard, {})

	def async_build_from_clipboard(self, content_params):
		"""This is used as a GSourceFunc so it should return False."""
		self.get_active_image().try_close_tab()
		if self.build_image_from_clipboard():
			self.switch_to(self.active_tool_id)
		return False

	def build_image_from_clipboard(self, *args):
		"""Open a new tab with the image in the clipboard. If the clipboard is
		empty, and there is no existing tab, a new blank image will be created
		to ensure the window can display the message.
		It returns `True` if an image (blank or not idc) has been created."""
		cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
		pixbuf = cb.wait_for_image()
		if pixbuf is None:
			self.reveal_message(_("The clipboard doesn't contain any image."), True)
			if self.notebook.get_current_page() < 0:
				# This condition means it's a new window.
				self.build_blank_image()
				return True
			return False
		else:
			self._build_new_tab(pixbuf=pixbuf)
			return True

	def build_image_from_selection(self, *args):
		"""Open a new tab with the image in the selection."""
		pixbuf = self.get_active_image().selection.get_pixbuf()
		self._build_new_tab(pixbuf=pixbuf)

	def build_new_from_file(self, gfile, check_duplicates=True):
		if check_duplicates:
			w, duplicate = self.app.has_image_opened(gfile.get_path())
			if duplicate is not None and not self.confirm_open_twice(gfile):
				w.notebook.set_current_page(duplicate)
				return
		self._build_new_tab(gfile=gfile)

	def _build_new_tab(self, gfile=None, pixbuf=None, \
		                     width=200, height=200, \
		                     background_rgba=[0.5, 0.5, 0.5, 0.5]):
		"""Open a new tab with an optional file to load in it, or directly a
		pixbuf, or the color and dimensions of a blank tab."""

		new_image = DrImage(self)
		self.notebook.append_page(new_image, new_image.build_tab_widget())
		self.notebook.child_set_property(new_image, 'reorderable', True)
		if gfile is not None:
			new_image.try_load_file(gfile)
		elif pixbuf is not None:
			new_image.try_load_pixbuf(pixbuf)
			# TODO dans l'idéal si on passe par les actions du clipboard, on
			# devrait n'ouvrir que si il n'y a pas de fenêtre, et ouvrir un truc
			# respectant les settings, plutôt qu'un petit pixbuf gris
		else:
			new_image.init_background(width, height, background_rgba)
		self._update_tabs_visibility()
		self.notebook.set_current_page(self.notebook.get_n_pages()-1)

	def confirm_open_twice(self, gfile):
		image_name = gfile.get_path().split('/')[-1]
		dialog = DrMessageDialog(self)
		# Context: %s is a file name
		label = _("The file %s is already opened")
		dialog.add_string(label % image_name)
		# Context: the user would click here to confirm they want to open the
		# same file twice
		open_again_id = dialog.set_action(_("Open again"), None)
		switch_to_id = dialog.set_action(_("Switch to this image"), \
		                                               'suggested-action', True)
		result = dialog.run()
		dialog.destroy()
		return result == open_again_id

	def on_active_tab_changed(self, *args):
		if not self._is_tools_initialisation_finished:
			return
		self.switch_to(self.active_tool_id, args[1])
		# print("changement d'image")
		self.update_picture_title(args[1].update_title())
		self.minimap.set_zoom_label(args[1].zoom_level * 100)
		args[1].update_image_wide_actions()
		# On devrait être moins bourrin et conserver la sélection # TODO ?

	def update_tabs_menu_section(self, *args):
		action = self.lookup_action('active_tab')
		section = self._get_menubar_item([[True, 2], [False, 1]])
		section.remove_all()
		for page in self.notebook.get_children():
			tab_title = page.update_title()
			tab_index = self.notebook.page_num(page)
			section.append(tab_title, 'win.active_tab(' + str(tab_index) + ')')

	def action_tab_left(self, *args):
		# XXX (un)availability of this action
		current_page = self.notebook.get_current_page()
		if current_page == 0:
			self.notebook.set_current_page(self.notebook.get_n_pages() - 1)
		else:
			self.notebook.set_current_page(current_page - 1)

	def action_tab_right(self, *args):
		# XXX (un)availability of this action
		current_page = self.notebook.get_current_page()
		if current_page == self.notebook.get_n_pages() - 1:
			self.notebook.set_current_page(0)
		else:
			self.notebook.set_current_page(current_page + 1)

	def close_tab(self, tab):
		"""Close a tab (after asking to save if needed)."""
		index = self.notebook.page_num(tab)
		if not self.notebook.get_nth_page(index).is_saved():
			self.notebook.set_current_page(index)
			is_saved = self.saving_manager.confirm_save_modifs()
			if not is_saved:
				return False
		self.notebook.remove_page(index)
		self._update_tabs_visibility()
		return True

	def action_close_tab(self, *args):
		if self.notebook.get_n_pages() > 1:
			self.get_active_image().try_close_tab()
		else:
			self.close()

	def action_close_window(self, *args):
		self.close()

	def on_close(self, *args):
		"""Event callback when trying to close a window. It saves/closes each
		tab and saves the current window settings in order to restore them.
		Returns `False` on success, `True` otherwise."""
		while self.notebook.get_n_pages() != 0:
			if not self.get_active_image().try_close_tab():
				return True

		self._decorations.remove_from_ui()
		self.options_manager.persist_tools_options()
		self.gsettings.set_string('last-active-tool', self.active_tool_id)
		self.gsettings.set_boolean('maximized', self.is_maximized())
		return False

	############################################################################
	# GENERAL PURPOSE METHODS ##################################################

	def connect_signals(self):
		# Closing the info bar
		self.info_bar.connect('close', self._hide_message)
		self.info_bar.connect('response', self._hide_message)

		# Closing the window
		self.connect('delete-event', self.on_close)

		# Resizing the window
		self.connect('configure-event', self._adapt_to_window_size)

		# When a setting changes
		self.gsettings.connect('changed::show-labels', self.on_show_labels_setting_changed)
		self.gsettings.connect('changed::deco-type', self.on_layout_changed)
		self.gsettings.connect('changed::big-icons', self.on_icon_size_changed)
		self.gsettings.connect('changed::preview-size', self.show_info_settings)
		# devel-only' isn't connected because it's auto-updated to False
		self.gsettings.connect('changed::disabled-tools', self.show_info_settings)
		self.gsettings.connect('changed::dark-theme-variant', self._update_theme_variant)
		# Other settings are connected in DrImage

		# What happens when the active image change
		self.notebook.connect('switch-page', self.on_active_tab_changed)

		# Change the active tool using "alt" mnemonics
		self.connect('key-press-event', self._check_for_alt_key)
		self.connect('key-release-event', self._check_for_alt_key)

		# Managing drag-and-drop
		if self.app.runs_in_sandbox:
			return # no dnd with actual file paths in the flatpak sandbox
			# XXX i could test if the app has home:ro permissions instead? eg to
			# help during development (where builder does have the permission)
		self.notebook.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.MOVE)
		self.notebook.connect('drag-data-received', self.on_data_dropped)
		self.notebook.drag_dest_add_uri_targets()
		# because drag_dest_add_image_targets doesn't work for files

	def _get_menubar_item(self, path_description):
		"""Get an item of the app-wide menubar. The `path_description` object
		is an array of [boolean, int] couples. The boolean means if we're
		looking for a submenu, the int is an index."""
		current_item = self.app.get_menubar()
		for item in path_description:
			if item[0]:
				link_type = Gio.MENU_LINK_SUBMENU
			else:
				link_type = Gio.MENU_LINK_SECTION
			current_item = current_item.get_item_link(item[1], link_type)
		return current_item

	def add_action_simple(self, action_name, callback, shortcuts=[]):
		"""Convenient wrapper method adding a stateless action to the window. It
		will be named 'action_name' (string) and activating the action will
		trigger the method 'callback'."""
		action = Gio.SimpleAction.new(action_name, None)
		action.connect('activate', callback)
		self.add_action(action)
		self.app.set_accels_for_action('win.' + action_name, shortcuts)

	def add_action_boolean(self, action_name, default, callback):
		"""Convenient wrapper method adding a stateful action to the window. It
		will be named 'action_name' (string), be created with the state 'default'
		(boolean), and activating the action will trigger the method 'callback'."""
		action = Gio.SimpleAction().new_stateful(action_name, None, \
		                                      GLib.Variant.new_boolean(default))
		action.connect('change-state', callback)
		self.add_action(action)

	def add_action_enum(self, action_name, default, callback):
		"""Convenient wrapper method adding a stateful action to the window. It
		will be named 'action_name' (string), be created with the state 'default'
		(string), and changing the active target of the action will trigger the
		method 'callback'."""
		action = Gio.SimpleAction().new_stateful(action_name, \
		            GLib.VariantType.new('s'), GLib.Variant.new_string(default))
		action.connect('change-state', callback)
		self.add_action(action)

	def add_all_win_actions(self):
		"""This doesn't add all window-wide GioActions, but only the GioActions
		which are here "by default", independently of any tool."""

		self.add_action_simple('main_menu', self.action_main_menu, ['F10'])
		self.add_action_simple('options_menu', self.action_options_menu, ['<Shift>F10'])

		action = Gio.PropertyAction.new('show-menubar', self, 'show-menubar')
		self.add_action(action)
		self.app.set_accels_for_action('win.show-menubar', ['<Ctrl>F2'])

		self.add_action_boolean('toggle_preview', False, self.action_toggle_preview)
		self.app.set_accels_for_action('win.toggle_preview', ['<Ctrl>m'])

		dark_variant = self.gsettings.get_boolean('dark-theme-variant')
		self.add_action_boolean('dark-variant', dark_variant, self.action_dark_theme)

		show_labels = self.gsettings.get_boolean('show-labels')
		self.add_action_boolean('show_labels', show_labels, self.action_show_labels)
		self.app.set_accels_for_action('win.show_labels', ['F9'])

		self.add_action_boolean('hide_controls', False, self.action_hide_controls)
		self.app.set_accels_for_action('win.hide_controls', ['F8'])
		self.lookup_action('hide_controls').set_enabled(False)

		self.add_action_boolean('fullscreen', False, self.action_fullscreen)
		self.app.set_accels_for_action('win.fullscreen', ['F11'])

		self.add_action_simple('reset_canvas', self.action_reset, ['<Ctrl>BackSpace'])
		self.add_action_simple('reload_file', self.action_reload, ['<Ctrl>r'])
		self.add_action_simple('properties', self.action_properties)
		self.add_action_simple('unfullscreen', self.action_unfullscreen)

		self.add_action_simple('go_up', self.action_go_up, ['<Ctrl>Up'])
		self.add_action_simple('go_down', self.action_go_down, ['<Ctrl>Down'])
		self.add_action_simple('go_left', self.action_go_left, ['<Ctrl>Left'])
		self.add_action_simple('go_right', self.action_go_right, ['<Ctrl>Right'])

		self.add_action_simple('go_top', self.action_go_top, \
		                                  ['<Ctrl>Page_Up', '<Ctrl>KP_Page_Up'])
		self.add_action_simple('go_bottom', self.action_go_bottom, \
		                              ['<Ctrl>Page_Down', '<Ctrl>KP_Page_Down'])
		self.add_action_simple('go_first', self.action_go_first, \
		                                        ['<Ctrl>Home', '<Ctrl>KP_Home'])
		self.add_action_simple('go_last', self.action_go_last, \
		                                          ['<Ctrl>End', '<Ctrl>KP_End'])

		self.add_action_simple('zoom_in', self.action_zoom_in, \
		                                         ['<Ctrl>plus', '<Ctrl>KP_Add'])
		self.add_action_simple('zoom_out', self.action_zoom_out, \
		                                   ['<Ctrl>minus', '<Ctrl>KP_Subtract'])
		self.add_action_simple('zoom_100', self.action_zoom_100, ['<Ctrl>1', '<Ctrl>KP_1'])
		self.add_action_simple('zoom_opti', self.action_zoom_opti, ['<Ctrl>0', '<Ctrl>KP_0'])
		self.add_action_simple('zoom_max', self.action_zoom_max, ['<Ctrl><Shift>plus', '<Ctrl><Shift>KP_Add'])

		self.add_action_simple('new_tab', self.build_blank_image, ['<Ctrl>t'])
		self.add_action_simple('new_tab_custom', self.build_new_custom)
		self.add_action_simple('new_tab_selection', \
		                    self.build_image_from_selection, ['<Ctrl><Shift>t'])
		self.add_action_simple('new_tab_clipboard', \
		                    self.build_image_from_clipboard, ['<Ctrl><Shift>v'])
		self.add_action_simple('open', self.action_open, ['<Ctrl>o'])
		self.add_action_simple('tab_right', self.action_tab_right, ['<Ctrl>Tab'])
		self.add_action_simple('tab_left', self.action_tab_left, ['<Ctrl><Shift>Tab'])
		self.add_action_simple('close_tab', self.action_close_tab, ['<Ctrl>w'])
		self.add_action_simple('close', self.action_close_window)

		self.add_action_simple('undo', self.action_undo, ['<Ctrl>z'])
		self.add_action_simple('redo', self.action_redo, ['<Ctrl><Shift>z'])

		self.add_action_simple('save', self.action_save, ['<Ctrl>s'])
		self.add_action_simple('save_alphaless', self.action_save_alphaless)
		self.add_action_simple('save_as', self.action_save_as, ['<Ctrl><Shift>s'])
		self.add_action_simple('export_as', self.action_export_as)
		self.add_action_simple('to_clipboard', self.action_export_cb, ['<Ctrl><Shift>c'])
		self.add_action_simple('print', self.action_print)

		self.add_action_simple('import', self.action_import, ['<Ctrl>i'])
		self.add_action_simple('paste', self.action_paste, ['<Ctrl>v'])
		self.add_action_simple('select_all', self.action_select_all, ['<Ctrl>a'])
		self.add_action_simple('unselect', self.action_unselect, ['<Ctrl><Shift>a'])
		#self.add_action_simple('selection_invert', self.action_selection_invert)
		self.add_action_simple('selection_cut', self.action_cut, ['<Ctrl>x'])
		self.add_action_simple('selection_copy', self.action_copy, ['<Ctrl>c'])
		self.add_action_simple('selection_delete', self.action_delete, ['Delete'])

		self.add_action_simple('selection_export', self.action_selection_export)
		self.add_action_simple('selection-replace-canvas', \
		                                   self.action_selection_replace_canvas)
		self.add_action_simple('selection-expand-canvas', \
		                                    self.action_selection_expand_canvas)

		self.add_action_simple('back_to_previous', self.back_to_previous, ['<Ctrl>b'])
		self.add_action_simple('apply_transform', self.action_apply_transform, ['<Ctrl>Return'])
		self.add_action_simple('cancel_transform', self.action_cancel_transform)

		self.add_action_enum('active_tool', DEFAULT_TOOL_ID, self.on_change_active_tool)

		self.add_action_simple('main_color', self.action_color1, ['<Ctrl><Shift>l'])
		self.add_action_simple('secondary_color', self.action_color2, ['<Ctrl><Shift>r'])
		self.add_action_simple('exchange_color', self.exchange_colors, ['<Ctrl>e'])

		editor = self.gsettings.get_boolean('direct-color-edit')
		self.app.add_action_boolean('use_editor', editor, self.action_use_editor)

		self.add_action_simple('size_more', self.action_size_more, ['<Ctrl><Shift>Up'])
		self.add_action_simple('size_less', self.action_size_less, ['<Ctrl><Shift>Down'])

		if self.devel_mode:
			self.add_action_simple('restore_pixbuf', self.action_restore)
			self.add_action_simple('rebuild_from_histo', self.action_rebuild)
			self.add_action_simple('get_values', self.action_getvalues, ['<Ctrl>g'])
			self.add_action_boolean('track_framerate', False, self.action_fsp)

		action = Gio.PropertyAction.new('active_tab', self.notebook, 'page')
		self.add_action(action)

	def set_cursor(self, is_custom):
		"""Called by the tools at various occasions, this method updates the
		mouse cursor according to `is_custom` (if False, use the default cursor)
		and the active tool `cursor_name` attribute."""
		if is_custom:
			name = self.active_tool().cursor_name
		else:
			name = 'default'
		cursor = Gdk.Cursor.new_from_name(Gdk.Display.get_default(), name)
		self.get_window().set_cursor(cursor)

	############################################################################
	# WINDOW DECORATIONS AND LAYOUTS ###########################################

	def on_layout_changed(self, *args):
		try:
			is_narrow = self._decorations.remove_from_ui()
			self._set_ui_bars()
			self._decorations.set_compact(is_narrow)
			self.update_picture_title()
			self.set_window_subtitles()
		except:
			pass # Closed windows are not actually deleted from the array kept
			# in main.py, so this can be called even after the window lost its
			# data (and it will fail).

	def show_info_settings(self, *args):
		"""This is executed when a setting changed but the method to apply it
		immediately in the current window doesn't exist."""
		self.reveal_message(_("Modifications will take effect in the next new window."))

	def update_picture_title(self, main_title=None):
		"""Set the window's title (regardless of the current UI bars), and the
		active tab title (indirectly)."""
		if main_title is None:
			main_title = self.get_active_image().update_title()
		self.update_tabs_menu_section()
		if self._decorations.set_title(main_title):
			self._decorations.force_update_titles()

	def set_window_subtitles(self, subtitles_list=None):
		"""Set the window's subtitles list (regardless of the current UI bars).
		Tools have to be initialized before calling this method, for now."""
		if subtitles_list is None:
			# TODO éviter ça en changeant les appels, et changer le commentaire
			subtitles_list = self.active_tool().get_editing_tips()
		if self._decorations.set_subtitles(subtitles_list):
			self._decorations.force_update_titles()

	def get_auto_decorations(self):
		"""Return the decorations setting based on the XDG_CURRENT_DESKTOP
		environment variable."""
		desktop_env = os.getenv('XDG_CURRENT_DESKTOP', 'GNOME')
		if 'GNOME' in desktop_env:
			return 'hg'
		elif 'Pantheon' in desktop_env:
			return 'he'
		elif 'Unity' in desktop_env:
			return 'tc'
		elif 'KDE' in desktop_env:
			return 'ts'
		elif 'Cinnamon' in desktop_env:
			return 'mts'
		elif desktop_env in ['MATE', 'XFCE', 'LXDE', 'LXQt']:
			return 'mtc'
		else:
			return 'hg' # Use the GNOME layout if the desktop is unknown,
		# because i don't know how the env variable is on mobile.

	def _set_ui_bars(self):
		"""Set the UI "bars" (headerbar, menubar, titlebar, toolbar, whatever)
		according to the user's preference, which by default is an empty string.
		In this case, an useful string is set by `get_auto_decorations()`."""
		self.has_good_width_limits = False

		# Remember the setting, so no need to restart this at each dialog.
		self.deco_layout = self.gsettings.get_string('deco-type')
		if self.deco_layout == '':
			self.deco_layout = self.get_auto_decorations()

		if self.deco_layout == 'hg':
			self._decorations = DrDecoManagerHeaderbar(False, self)
		elif self.deco_layout == 'he':
			self._decorations = DrDecoManagerHeaderbar(True, self)
		elif self.deco_layout == 'm':
			self._decorations = DrDecoManagerMenubar(self, True)
		elif 't' in self.deco_layout:
			symbolic = 's' in self.deco_layout
			menubar = 'm' in self.deco_layout
			self._decorations = DrDecoManagerToolbar(symbolic, menubar, self)
		else:
			self.gsettings.set_string('deco-type', '')
			self._set_ui_bars() # yes, recursion.

		if self.app.is_beta():
			self.get_style_context().add_class('devel')

	def action_main_menu(self, *args):
		self._decorations.toggle_menu()

	def action_options_menu(self, *args):
		"""This displays/hides the tool's options menu, and is implemented as an
		action to ease the accelerator (shift+f10)."""
		self.options_manager.toggle_menu()

	def _adapt_to_window_size(self, *args):
		"""Adapts the headerbar (if any) and the default bottom pane to the new
		window size. If the current bottom pane isn't the default one, this
		will call the tool method applying the new size to the tool pane."""
		if not self.has_good_width_limits and self.get_allocated_width() > 700:
			self.options_manager.init_adaptability()
			self._decorations.init_adaptability()
			self.has_good_width_limits = True
		self._decorations.adapt_to_window_size()

		available_width = self.bottom_panes_box.get_allocated_width()
		if not self._is_tools_initialisation_finished:
			return # there is no active pane nor active image yet
		self.options_manager.adapt_to_window_size(available_width)

		# Check whether or not the scrollbars should now be displayed
		self.get_active_image().fake_scrollbar_update()

	def _update_tabs_visibility(self):
		controls_hidden = self.lookup_action('hide_controls').get_state()
		should_show = (self.notebook.get_n_pages() > 1) and not controls_hidden
		self.notebook.set_show_tabs(should_show)

	def action_dark_theme(self, *args):
		shall_be_dark = args[1]
		self.gsettings.set_boolean('dark-theme-variant', shall_be_dark)
		args[0].set_state(GLib.Variant.new_boolean(shall_be_dark))

	def _update_theme_variant(self, *args):
		key = 'gtk-application-prefer-dark-theme';
		use_dark_theme = self.gsettings.get_boolean('dark-theme-variant')
		Gtk.Settings.get_default().set_property(key, use_dark_theme)
		# XXX intriguant ce truc là ^ imho ça a changé, le night theme existe

	############################################################################
	# INFORMATION MESSAGES #####################################################

	def log_message(self, message):
		if self.devel_mode:
			print("Drawing: " + message)

	def reveal_action_report(self, message):
		"""Update the label and the visibility of the info bar. The 'report bug'
		action is suggested."""
		self.info_bar.set_visible(True)
		self.info_label.set_label(message)
		self.info_action.set_action_name('app.report-issue')
		self.info_action.set_label(_("Report a bug"))
		self.info_action.set_visible(True)
		self.log_message(message)

	def reveal_message(self, label, hide_afterwards=False):
		"""Update the label and the visibility of the info bar. No action is
		suggested, and the info bar may optionally be hidden after 4 seconds."""
		if hide_afterwards:
			GLib.timeout_add(4000, self.__hide_message_async, {'label': label})
		self.info_bar.set_visible(True)
		self.info_action.set_visible(False)
		self.info_label.set_label(label)
		self.log_message(label)

	def __hide_message_async(self, async_cb_data):
		"""This is used as a GSourceFunc so it should return False."""
		if async_cb_data['label'] == self.info_label.get_label():
			self._hide_message()
		# else the message has changed so it shouldn't be hidden now
		return False

	def _hide_message(self, *args):
		self.info_bar.set_visible(False)
		self.info_action.set_visible(False)
		self.info_label.set_label("")

	############################################################################
	# FULLSCREEN ###############################################################

	def action_unfullscreen(self, *args):
		"""Simple action, exiting fullscreen."""
		fs_action = self.lookup_action('fullscreen')
		fs_action.change_state(GLib.Variant.new_boolean(False))

	def action_fullscreen(self, *args):
		"""Boolean action, toggling fullscreen mode."""
		# XXX maybe track the window widget's 'window-state-event' to use this
		# https://lazka.github.io/pgi-docs/Gdk-3.0/flags.html#Gdk.WindowState.FULLSCREEN
		shall_fullscreen = args[1]
		if shall_fullscreen:
			self.fullscreen()
			self.reveal_message(_("Middle-click or press F8 to show/hide controls.") + \
			                     " " + _("Press F11 to exit fullscreen."), True)
			# TODO find a solution for touchscreens!!
		else:
			self.unfullscreen()
		self._set_controls_hidden(shall_fullscreen)
		self.lookup_action('hide_controls').set_enabled(shall_fullscreen)
		self.unfullscreen_btn.set_visible(shall_fullscreen)
		args[0].set_state(GLib.Variant.new_boolean(shall_fullscreen))

	def action_hide_controls(self, *args):
		"""Boolean action controlling the visibility of the UI elements such as
		the tools list, the bottom pane, the menubar/toolbar if any. This can
		only (in theory) be used while in fullscreen."""
		controls_hidden = args[1]
		self.tools_flowbox.set_visible(not controls_hidden)
		if 't' in self.deco_layout:
			self.toolbar_box.set_visible(not controls_hidden)
		if 'm' in self.deco_layout:
			self.set_show_menubar(not controls_hidden)
		self.bottom_meta_box.set_visible(not controls_hidden)
		args[0].set_state(GLib.Variant.new_boolean(controls_hidden))
		self._update_tabs_visibility()

	def _set_controls_hidden(self, state):
		hc_action = self.lookup_action('hide_controls')
		hc_action.change_state(GLib.Variant.new_boolean(state))

	def on_middle_click(self):
		is_fullscreened = self.lookup_action('fullscreen').get_state()
		if is_fullscreened:
			hc_action = self.lookup_action('hide_controls')
			self._set_controls_hidden(not hc_action.get_state())
		else:
			self.options_manager.get_active_pane().middle_click_action()

	############################################################################
	# SIDE PANE (TOOLS) ########################################################

	def on_icon_size_changed(self, *args):
		for tool_id in self.tools:
			self.tools[tool_id].update_icon_size()

	def set_tools_labels_visibility(self, visible):
		"""Change the way tools are displayed in the side pane. Visible labels
		mean the tools will be arranged in a scrollable list of buttons, else
		they will be in an adaptative flowbox."""
		for tool_id in self.tools:
			self.tools[tool_id].set_show_label(visible)
		nb_tools = len(self.tools)
		parent_box = self.tools_flowbox.get_parent()
		if visible:
			self.tools_flowbox.set_min_children_per_line(nb_tools)
			if parent_box == self.tools_nonscrollable_box:
				self.tools_nonscrollable_box.remove(self.tools_flowbox)
				self.tools_scrollable_box.add(self.tools_flowbox)
		else:
			if parent_box == self.tools_scrollable_box:
				self.tools_scrollable_box.remove(self.tools_flowbox)
				self.tools_nonscrollable_box.add(self.tools_flowbox)
			nb_min = int( (nb_tools+(nb_tools % 3))/3 ) - 1
			self.tools_flowbox.set_min_children_per_line(nb_min)
		self.tools_flowbox.set_max_children_per_line(nb_tools)

	def on_show_labels_setting_changed(self, *args):
		# TODO https://lazka.github.io/pgi-docs/Gio-2.0/classes/Settings.html#Gio.Settings.create_action
		self.set_tools_labels_visibility(self.gsettings.get_boolean('show-labels'))

	def action_show_labels(self, *args):
		show_labels = not args[0].get_state()
		self.gsettings.set_boolean('show-labels', show_labels)
		args[0].set_state(GLib.Variant.new_boolean(show_labels))

	def _check_for_alt_key(self, *args):
		if not args[1].state | Gdk.ModifierType.MOD1_MASK == args[1].state:
			return
		is_press = args[1].type == Gdk.EventType.KEY_PRESS
		for tool_id in self.tools:
			self.tools[tool_id].show_only_mnemonics(is_press)

	############################################################################
	# TOOLS ####################################################################

	def on_change_active_tool(self, *args):
		"""Action callback, doing nothing in some situations, thus avoiding
		infinite loops. It sets the correct `tool_id` using adequate methods
		otherwise."""
		state_as_string = args[1].get_string()
		if state_as_string == args[0].get_state().get_string():
			return
		if self.tools[state_as_string].is_flowbox_child_selected():
			self.switch_to(state_as_string)
		else:
			self.tools[state_as_string].select_flowbox_child()

	def switch_to(self, new_tool_id, image_pointer=None):
		"""Switch from the current tool to `new_tool_id` and to the current
		image to `image_pointer` (`None` if the image is not changing)."""
		self.pointer_to_current_page = None

		action = self.lookup_action('active_tool')
		action.set_state(GLib.Variant.new_string(new_tool_id))

		# Disable the formerly active tool
		self.former_tool_id = self.active_tool_id
		should_preserve_selection = self.tools[new_tool_id].accept_selection
		self.former_tool().give_back_control(should_preserve_selection)
		self.former_tool().on_tool_unselected()
		self.get_active_image().selection.hide_popovers()

		self.pointer_to_current_page = image_pointer

		# Enable the newly selected tool
		self.get_active_image().update()
		self.active_tool_id = new_tool_id
		self.active_tool().on_tool_selected()
		self._update_bottom_pane()
		self.get_active_image().update_actions_state()
		self.set_window_subtitles()

		self.pointer_to_current_page = None

	def _update_bottom_pane(self):
		"""Show the correct bottom pane, with the correct tool options menu."""
		try:
			self.options_manager.try_enable_pane(self.active_tool().pane_id)
			self.options_manager.update_pane(self.active_tool())
			self._build_options_menu()
			self._adapt_to_window_size()
		except Exception as e:
			self.reveal_message(_("Error loading the bottom pane for the " + \
			        "tool '%s', please report this bug.") % self.active_tool_id)
			print(e)

	def active_tool(self):
		return self.tools[self.active_tool_id]

	def former_tool(self):
		return self.tools[self.former_tool_id]

	def back_to_previous(self, *args):
		"""Switch back to the previously active tool, if it's different from the
		current one."""
		if self.former_tool_id == self.active_tool_id:
			self.force_selection()
			# avoid cases where applying a transform tool keeps the tool active
		else:
			self.tools[self.former_tool_id].select_flowbox_child()

	def _build_options_menu(self):
		"""Build the active tool's option menus.
		The first menu is the popover from the bottom bar. It can be built from
		a Gio.MenuModel, or it can contain any widget.
		The second menu is build from a Gio.MenuModel and is in the menubar (not
		available with all layouts)."""
		widget = self.active_tool().get_options_widget()
		model = self.active_tool().get_options_model()
		label = self.active_tool().get_options_label()
		if model is None:
			self.app.get_menubar().remove(5)
			item = Gio.MenuItem()
			item.set_label(label)
			item.set_action_and_target_value('win.PLACEHOLDER', None)
			self.app.get_menubar().insert_item(5, item)
		else:
			self.app.get_menubar().remove(5)
			self.app.get_menubar().insert_submenu(5, label, model)
		pane = self.options_manager.get_active_pane()
		pane.build_options_menu(widget, model, label)

	def action_use_editor(self, *args):
		use_editor = not args[0].get_state()
		self.gsettings.set_boolean('direct-color-edit', use_editor)
		args[0].set_state(GLib.Variant.new_boolean(use_editor))
		self.options_manager.set_palette_setting(use_editor)

	def exchange_colors(self, *args):
		self.options_manager.get_classic_tools_pane().middle_click_action()

	def action_color1(self, *args):
		if self.active_tool().use_color:
			self.options_manager.left_color_btn().open()

	def action_color2(self, *args):
		if self.active_tool().use_color:
			self.options_manager.right_color_btn().open()

	def action_size_more(self, *args):
		self.options_manager.update_tool_width(1)

	def action_size_less(self, *args):
		self.options_manager.update_tool_width(-1)

	############################################################################
	# IMAGE FILES MANAGEMENT ###################################################

	def action_properties(self, *args):
		"""Display the properties dialog for the current image. This could be
		done here but it's done in DrImage to have a satisfying UML diagram."""
		self.get_active_image().show_properties()

	def action_fsp(self, *args):
		"""Development only: tracks and displays the framerate, thus it helps
		debugging how Gdk/cairo draws on the widget."""
		self.should_track_framerate = not self.should_track_framerate
		for img in self.notebook.get_children():
			img.reset_fps_counter()
		args[0].set_state(GLib.Variant.new_boolean(self.should_track_framerate))

	def get_active_image(self):
		if self.pointer_to_current_page is None:
			return self.notebook.get_nth_page(self.notebook.get_current_page())
		else:
			return self.pointer_to_current_page

	def action_reset(self, *args):
		self.get_active_image().reset_to_initial_pixbuf()

	def action_reload(self, *args):
		self.get_active_image().reload_from_disk()

	def action_open(self, *args):
		"""Handle the result of an "open" file chooser dialog, and open it in
		the current tab, or in a new one, or in a new window. The decision is
		made depending on what's in the current tab, and (if any doubt)
		according to the user explicit decision."""
		gfile = self.file_chooser_open()
		if gfile is None:
			return
		else:
			file_name = gfile.get_path().split('/')[-1]
			self.reveal_message(_("Loading %s") % file_name)
		if self.get_active_image().should_replace():
			# XXX ne marche pas de ouf en pratique ^
			# If the current image is just a blank, unmodified canvas.
			self._try_load_file(gfile)
		else:
			# it makes more sense to ask *if* the user want to open it BEFORE
			# asking *where* to open it
			w, duplicate = self.app.has_image_opened(gfile.get_path())
			if duplicate is not None and not self.confirm_open_twice(gfile):
				w.notebook.set_current_page(duplicate)
				self._hide_message()
				return

			dialog = DrMessageDialog(self)
			new_tab_id = dialog.set_action(_("New Tab"), None, True)
			new_window_id = dialog.set_action(_("New Window"), None)
			discard_id = dialog.set_action(_("Discard changes"), 'destructive-action')
			if not self.get_active_image().is_saved():
				# Context: %s will be replaced by the name of a file.
				dialog.add_string(_("There are unsaved modifications to %s.") % \
				             self.get_active_image().get_filename_for_display())
			# Context: %s will be replaced by the name of a file.
			dialog.add_string(_("Where do you want to open %s?") % file_name)
			result = dialog.run()
			dialog.destroy()
			if result == new_tab_id:
				self.build_new_from_file(gfile, False)
			elif result == discard_id:
				self._try_load_file(gfile, False)
			elif result == new_window_id:
				self.app.open_window_with_content(gfile, False, False)
		self._hide_message()

	def file_chooser_open(self, *args):
		"""Opens an "open" file chooser dialog, and return a GioFile or None."""
		gfile = None
		file_chooser = Gtk.FileChooserNative.new(_("Open a picture"), self,
		                     Gtk.FileChooserAction.OPEN, _("Open"), _("Cancel"))
		utilities_add_filechooser_filters(file_chooser)
		response = file_chooser.run()
		if response == Gtk.ResponseType.ACCEPT:
			gfile = file_chooser.get_file()
		file_chooser.destroy()
		return gfile

	def on_data_dropped(self, widget, drag_context, x, y, data, info, time):
		"""Signal callback: when files are dropped on `self.notebook`, a message
		dialog is shown, asking if the user prefers to open them (one new tab
		per image), or to import them (it will only import the first), or to
		cancel (if the user dropped mistakenly)."""
		dialog = DrMessageDialog(self)
		cancel_id = dialog.set_action(_("Cancel"), None)
		open_id = dialog.set_action(_("Open"), None)
		import_id = dialog.set_action(_("Import"), None, True)

		uris = data.get_uris()
		gfiles = []
		for uri in uris:
			try:
				gfile = Gio.File.new_for_uri(uri)
				is_valid_image, error_msg = utilities_gfile_is_image(gfile)
			except Exception as excp:
				is_valid_image = False
				error_msg = excp.message
			if is_valid_image:
				gfiles.append(gfile)
			else:
				self.reveal_message(error_msg)

		if len(gfiles) == 0:
			return
		elif len(gfiles) == 1:
			label = gfiles[0].get_path().split('/')[-1]
		else:
			# Context for translation:
			# "What do you want to do with *these files*?"
			label = _("these files")
		# Context: %s will be replaced by the name of a file. The possible
		# answers are "cancel", "open", and "import"
		dialog.add_string(_("What do you want to do with %s?") % label)
		result = dialog.run()
		dialog.destroy()

		if result == open_id:
			for f in gfiles:
				self._build_new_tab(gfile=f)
		elif result == import_id:
			self.import_from_path(gfiles[0].get_path())

	def _try_load_file(self, gfile, check_duplicates=True):
		if gfile is None:
			return
		if check_duplicates:
			w, duplicate = self.app.has_image_opened(gfile.get_path())
			if w is not None and not self.confirm_open_twice(gfile):
				w.notebook.set_current_page(duplicate)
				return

		self.get_active_image().try_load_file(gfile)
		self.update_picture_title()

	def has_image_opened(self, file_path):
		for tab in self.notebook.get_children():
			if tab.get_file_path() == file_path:
				return self.notebook.page_num(tab)
		return None

	def action_save(self, *args):
		"""Try to save the active image, and return True if the image has been
		successfully saved."""
		return self.saving_manager.save_current_image(False, False, False, True)

	def action_save_as(self, *args):
		return self.saving_manager.save_current_image(False, True, False, True)

	def action_save_alphaless(self, *args):
		return self.saving_manager.save_current_image(False, False, False, False)

	def action_export_as(self, *args):
		return self.saving_manager.save_current_image(True, True, False, True)

	def action_print(self, *args):
		pixbuf = self.get_active_image().main_pixbuf
		self.printing_manager.print_pixbuf(pixbuf)

	def action_export_cb(self, *args):
		cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
		cb.set_image(self.get_active_image().main_pixbuf)
		self.reveal_message(_("Image copied to clipboard"), True)

	############################################################################
	# SELECTION MANAGEMENT #####################################################

	def action_getvalues(self, *args):
		"""Development only: helps debugging the selection."""
		self.get_active_image().selection.print_values()

	def action_select_all(self, *args):
		self.force_selection()
		self.get_selection_tool().select_all()

	def action_unselect(self, *args):
		self.get_selection_tool().give_back_control(False)

	def action_cut(self, *args):
		self.copy_operation()
		self.action_delete()

	def action_copy(self, *args):
		self.copy_operation()

	def copy_operation(self):
		cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
		cb.set_image(self.get_active_image().selection.get_pixbuf())

	def action_delete(self, *args):
		self.get_selection_tool().delete_selection()

	def action_paste(self, *args):
		"""By default, this action pastes an image, but if there is no image in
		the clipboard, it will paste text using the text tool. Once the text
		tool is active, this action is disabled to not interfere with the default
		behavior of ctrl+v provided by the GTK text entry."""
		cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
		pixbuf = cb.wait_for_image()
		if pixbuf is not None:
			self.force_selection()
			self.get_selection_tool().import_selection(pixbuf)
		else:
			string = cb.wait_for_text()
			if string is not None:
				self.tools['text'].force_text_tool(string)

	def action_import(self, *args):
		"""Handle the result of an 'open' file chooser dialog. It will then try
		to import it as the selection."""
		file_chooser = Gtk.FileChooserNative.new(_("Import a picture"), self,
		                   Gtk.FileChooserAction.OPEN, _("Import"), _("Cancel"))
		utilities_add_filechooser_filters(file_chooser)
		response = file_chooser.run()
		if response == Gtk.ResponseType.ACCEPT:
			self.import_from_path(file_chooser.get_filename())
		file_chooser.destroy()

	def import_from_path(self, file_path):
		"""Import a file as the selection pixbuf. Called by the 'win.import'
		action or when an image is imported by drag-and-drop."""
		self.force_selection()
		pixbuf = GdkPixbuf.Pixbuf.new_from_file(file_path)
		self.get_selection_tool().import_selection(pixbuf)

	def action_selection_export(self, *args):
		return self.saving_manager.save_current_image(True, True, True, True)

	def action_selection_replace_canvas(self, *args):
		self.get_selection_tool().replace_canvas()

	def action_selection_expand_canvas(self, *args):
		crop_tool = self.tools['crop']
		operation = crop_tool.build_selection_fit_operation()
		crop_tool.apply_operation(operation) # calling this here isn't elegant

	def action_selection_invert(self, *args):
		self.get_selection_tool().invert_selection()

	def get_selection_tool(self):
		if 'rect_select' in self.tools:
			return self.tools['rect_select']
		elif 'free_select' in self.tools:
			return self.tools['free_select']
		elif 'color_select' in self.tools:
			return self.tools['color_select']
		else:
			self.reveal_action_report(_("Required tool is not available"))
			return self.active_tool()

	def force_selection(self):
		self.get_selection_tool().select_flowbox_child()

	def action_apply_transform(self, *args):
		self.active_tool().on_apply_transform_tool_operation()

	def action_cancel_transform(self, *args):
		self.active_tool().on_cancel_transform_tool_operation()

	############################################################################
	# HISTORY MANAGEMENT #######################################################

	def action_undo(self, *args):
		# self.reveal_message(_("Undoing…"))
		self.get_active_image().try_undo()
		# self.log_message('finished undoing')
		# self._hide_message()

	def action_redo(self, *args):
		self.get_active_image().try_redo()

	def action_restore(self, *args):
		"""[Dev only] show the last saved pixbuf on the canvas."""
		self.get_active_image().use_stable_pixbuf()
		self.get_active_image().update()

	def action_rebuild(self, *args):
		"""[Dev only] rebuild the image according to the history content."""
		self.get_active_image()._history._rebuild_from_history()

	def update_history_actions_labels(self, undo_label, redo_label):
		self._decorations.set_undo_label(undo_label)
		self._decorations.set_redo_label(redo_label)

	############################################################################
	# PREVIEW, NAVIGATION AND ZOOM ACTIONS #####################################

	def action_go_up(self, *args):
		self.get_active_image().add_deltas(0, -1, 100)

	def action_go_down(self, *args):
		self.get_active_image().add_deltas(0, 1, 100)

	def action_go_left(self, *args):
		self.get_active_image().add_deltas(-1, 0, 100)

	def action_go_right(self, *args):
		self.get_active_image().add_deltas(1, 0, 100)

	def action_go_top(self, *args):
		self.get_active_image().reset_deltas(0, -1)

	def action_go_bottom(self, *args):
		self.get_active_image().reset_deltas(0, 1)

	def action_go_first(self, *args):
		self.get_active_image().reset_deltas(-1, 0)

	def action_go_last(self, *args):
		self.get_active_image().reset_deltas(1, 0)

	def action_toggle_preview(self, *args):
		"""Action callback, showing or hiding the "minimap" preview popover."""
		preview_visible = not args[0].get_state()
		if preview_visible:
			self.minimap.popup()
			self.minimap.update_content()
		else:
			self.minimap.popdown()
		args[0].set_state(GLib.Variant.new_boolean(preview_visible))

	def action_zoom_in(self, *args):
		self.get_active_image().inc_zoom_level(25)

	def action_zoom_out(self, *args):
		self.get_active_image().inc_zoom_level(-25)

	def action_zoom_100(self, *args):
		self.get_active_image().set_zoom_level(100)

	def action_zoom_opti(self, *args):
		self.get_active_image().set_opti_zoom_level()

	def action_zoom_max(self, *args):
		self.get_active_image().set_zoom_level(2000)
		# It has to be done twice, because the code of the minimap is shit
		self.get_active_image().set_zoom_level(2000)

	############################################################################
################################################################################




# select_rect.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .abstract_select import AbstractSelectionTool
from .utilities_overlay import utilities_show_overlay_on_context

class ToolRectSelect(AbstractSelectionTool):
	__gtype_name__ = 'ToolRectSelect'

	def __init__(self, window, **kwargs):
		super().__init__('rect_select', _("Rectangle selection"), 'tool-select-rect-symbolic', window)

	def get_tooltip(self, event_x, event_y, motion_behavior):
		if motion_behavior != 1:
			return None # no line is being drawn

		delta_x = abs(self.x_press - event_x)
		delta_y = abs(self.y_press - event_y)
		line1 = _("Width: %spx") % str(delta_x)
		line2 = _("Height: %spx") % str(delta_y)
		return line1 + "\n" + line2

	############################################################################

	def press_define(self, event_x, event_y):
		pass

	def motion_define(self, event_x, event_y, render):
		if not render:
			return
		self._build_rectangle_path(self.x_press, self.y_press, event_x, event_y)
		self.restore_pixbuf()
		rect = self.get_selection().get_future_path()
		if rect is not None:
			ccontext = self.get_context()
			thickness = self.get_overlay_thickness()
			utilities_show_overlay_on_context(ccontext, rect, thickness)

	def release_define(self, surface, event_x, event_y):
		self._build_rectangle_path(self.x_press, self.y_press, event_x, event_y)
		self.operation_type = 'op-define'
		operation = self.build_operation()
		self.apply_operation(operation)
		# self.get_selection().show_popover()

	############################################################################
################################################################################



# tool_crop.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk, GdkPixbuf
from .abstract_transform_tool import AbstractCanvasTool
from .optionsbar_crop import OptionsBarCrop
from .utilities_overlay import utilities_show_handles_on_context
from .utilities_colors import utilities_rgba_to_hexadecimal, \
                              utilities_gdk_rgba_to_color_array

class ToolCrop(AbstractCanvasTool):
	__gtype_name__ = 'ToolCrop'

	def __init__(self, window):
		super().__init__('crop', _("Crop"), 'tool-crop-symbolic', window)
		self.cursor_name = 'not-allowed'
		self.x_press = self.x_motion = 0
		self.y_press = self.y_motion = 0
		self._unclicked = True # locking operation execution to avoid inf. loops
		self.add_tool_action_enum('crop-expand', 'initial')

	def try_build_pane(self):
		self.pane_id = 'crop'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		bar = OptionsBarCrop()
		self.height_btn = bar.height_btn
		self.width_btn = bar.width_btn
		self.width_btn.connect('value-changed', self.on_width_changed)
		self.height_btn.connect('value-changed', self.on_height_changed)
		return bar

	def get_options_label(self):
		# The options of the "crop" tool are all about how to expand the canvas,
		# hence this label.
		return _("Expanding options")

	def get_editing_tips(self):
		label_direction = _("The sides you'll crop are hinted by the mouse pointer")

		label_modifiers = None
		if self.apply_to_selection:
			label_action = _("Cropping the selection")
			label_confirm = None
		else:
			label_action = _("Cropping or expanding the canvas")
			label_confirm = self.label + " - " + \
			                         _("Don't forget to confirm the operation!")
			if not self.get_image().get_mouse_is_pressed():
				label_modifiers = _("Press <Alt>, <Shift>, or both, to " + \
				                      "quickly change the 'expand with' option")

		full_list = [label_action, label_direction, label_confirm, label_modifiers]
		return list(filter(None, full_list))

	############################################################################

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self._x = 0
		self._y = 0
		if self.apply_to_selection:
			self._init_if_selection()
		else:
			self._init_if_main()
		self._update_expansion_rgba()
		self.width_btn.set_value(self._original_width)
		self.height_btn.set_value(self._original_height)
		self.set_action_sensitivity('crop-expand', not self.apply_to_selection)
		self.build_and_do_op()

	def _init_if_selection(self):
		self._original_width = self.get_selection_pixbuf().get_width()
		self._original_height = self.get_selection_pixbuf().get_height()
		self.width_btn.set_range(1, self._original_width)
		self.height_btn.set_range(1, self._original_height)

	def _init_if_main(self):
		self._original_width = self.get_image().get_pixbuf_width()
		self._original_height = self.get_image().get_pixbuf_height()
		self.width_btn.set_range(1, 10 * self._original_width)
		self.height_btn.set_range(1, 10 * self._original_height)

	############################################################################

	def _get_width(self):
		return self.width_btn.get_value_as_int()

	def _get_height(self):
		return self.height_btn.get_value_as_int()

	def on_width_changed(self, *args):
		if self._unclicked:
			self.build_and_do_op()

	def on_height_changed(self, *args):
		if self._unclicked:
			self.build_and_do_op()

	def move_north(self, delta):
		self.height_btn.set_value(self.height_btn.get_value() - delta)
		self._y = self._y + delta

	def move_south(self, delta):
		self.height_btn.set_value(self.height_btn.get_value() + delta)

	def move_east(self, delta):
		self.width_btn.set_value(self.width_btn.get_value() + delta)

	def move_west(self, delta):
		self.width_btn.set_value(self.width_btn.get_value() - delta)
		self._x = self._x + delta

	############################################################################

	def on_unclicked_motion_on_area(self, event, surface):
		self.set_directional_cursor(event.x, event.y)

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.x_press = self.x_motion = event_x
		self.y_press = self.y_motion = event_y
		self._unclicked = False
		self._update_expansion_rgba(event.button)

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys and 'ALT' in self._modifier_keys:
			self._force_expansion_rgba('secondary', event.button)
		elif 'SHIFT' in self._modifier_keys:
			self._force_expansion_rgba('alpha')
		elif 'ALT' in self._modifier_keys:
			self._force_expansion_rgba('initial')

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		delta_x = event_x - self.x_motion
		delta_y = event_y - self.y_motion

		if self._directions == '':
			return
		if 'n' in self._directions:
			self.move_north(delta_y)
		if 's' in self._directions:
			self.move_south(delta_y)
		if 'w' in self._directions:
			self.move_west(delta_x)
		if 'e' in self._directions:
			self.move_east(delta_x)

		if self.apply_to_selection:
			self._x = max(0, self._x)
			self._y = max(0, self._y)
			width = min(self._original_width - self._x, self._get_width())
			height = min(self._original_height - self._y, self._get_height())
			self.width_btn.set_value(width)
			self.height_btn.set_value(height)

		self.x_motion = event_x
		self.y_motion = event_y
		# not adding the condition would be a better UX but slower to compute
		if render and not self.apply_to_selection:
			self.build_and_do_op()

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._unclicked = True
		self.build_and_do_op()
		self._scroll_to_end(event_x - self.x_press, event_y - self.y_press)
		self.window.set_cursor(False)

	############################################################################

	def on_draw_above(self, area, cairo_context):
		if self.apply_to_selection:
			x1 = int(self._x)
			y1 = int(self._y)
		else:
			x1 = 0
			y1 = 0
		x2 = x1 + self._get_width()
		y2 = y1 + self._get_height()
		x1, x2, y1, y2 = self.get_image().get_corrected_coords(x1, x2, y1, y2, \
		                                         self.apply_to_selection, False)
		if not self.apply_to_selection:
			self._draw_temp_pixbuf(cairo_context, x1, y1)
		thickness = self.get_overlay_thickness()
		utilities_show_handles_on_context(cairo_context, x1, x2, y1, y2, thickness)

	############################################################################

	def build_selection_fit_operation(self):
		"""Special way to build an operation, not from the present crop tool,
		but from the selection menu. The parameters are determined automatically
		from the state of the selection manager."""
		self._update_expansion_rgba()
		s = self.get_selection()
		new_x = min(0, s.selection_x)
		new_y = min(0, s.selection_y)

		s_width = max(0, s.selection_x) + s.selection_pixbuf.get_width()
		new_width = max(-1 * new_x + self.get_main_pixbuf().get_width(), s_width)

		s_height = max(0, s.selection_y) + s.selection_pixbuf.get_height()
		new_height = max(-1 * new_y + self.get_main_pixbuf().get_height(), s_height)

		operation = {
			'tool_id': self.id,
			'is_selection': False,
			'is_preview': False,
			'is_etf': True,
			'local_dx': int(new_x),
			'local_dy': int(new_y),
			'width': new_width,
			'height': new_height,
			'rgba': self._expansion_rgba
		}
		return operation

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'is_selection': self.apply_to_selection,
			'is_preview': True,
			'is_etf': False,
			'local_dx': int(self._x),
			'local_dy': int(self._y),
			'width': self._get_width(),
			'height': self._get_height(),
			'rgba': self._expansion_rgba
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		x = operation['local_dx']
		y = operation['local_dy']
		width = operation['width']
		height = operation['height']

		rgba_array = utilities_gdk_rgba_to_color_array(operation['rgba'])
		rgba_array[3] *= 255
		hexa_rgba = utilities_rgba_to_hexadecimal(*rgba_array)

		is_selection = operation['is_selection']
		if is_selection:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()
		self.get_image().set_temp_pixbuf(source_pixbuf.copy())
		self._crop_temp_pixbuf(x, y, width, height, is_selection, hexa_rgba)
		if operation['is_etf']:
			# Case of an "expand to fit" action
			s_pixbuf = self.get_selection_pixbuf()
			self.get_selection().update_from_transform_tool(s_pixbuf, -1 * x, -1 * y)
		self.common_end_operation(operation)

	def _crop_temp_pixbuf(self, x, y, width, height, is_selection, hexa_rgba):
		"""Crop and/or expand the temp pixbuf according to given parameters."""

		# Coordinates of the origin of the source pixbuf (temp_p)
		src_x = max(x, 0)
		src_y = max(y, 0)

		# Coordinates of the origin of the destination pixbuf (new_pixbuf)
		if is_selection:
			dest_x = 0
			dest_y = 0
		else:
			dest_x = max(-1 * x, 0)
			dest_y = max(-1 * y, 0)

		# Dotted lines == new sizes; plain lines == old sizes.
		#
		# If the origin has been cropped, `src` == the new coordinates, and
		# `dest` == 0 (in the considered direction x or y):
		#
		# dest____________________
		# |                       |
		# |  src------------------|
		# |   ⁝                   |
		# |___⁝___________________|
		#
		# With the selection, it's not possible to expand, so we're always in
		# this first case, with `dest` == 0.
		# Else (if it hasn't moved, or if it has been expanded) `src` has to be
		# 0 (because the coordinates GdkPixbuf will use can't be negative), and
		# `dest` is (-1 * the new coordinates):
		#
		# dest---------------------
		# ⁝                       ⁝
		# ⁝  src__________________⁝
		# ⁝   |                   |
		# ⁝___|___________________|
		#
		# The sign inversion is completely artificial (again: it's because the
		# coordinates GdkPixbuf will use can't be negative).

		# Initialisation of an EMPTY pixbuf with the wanted size and color
		new_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, \
		                                                          width, height)
		new_pixbuf.fill(hexa_rgba)

		temp_p = self.get_image().temp_pixbuf
		# The width/height we want (measured from the respective origins of the
		# `src` and the `dest` rectangles)
		min_w = min(width - dest_x, temp_p.get_width() - src_x)
		min_h = min(height - dest_y, temp_p.get_height() - src_y)

		# Copy an area of the source pixbuf `temp_p`; the area starts at `src_*`
		# and has the dimensions `min_*`. It's painted on the destination pixbuf
		# (`new_pixbuf`) starting at the coordinates `dest_*`.
		temp_p.copy_area(src_x, src_y, min_w, min_h, new_pixbuf, dest_x, dest_y)
		self.get_image().set_temp_pixbuf(new_pixbuf)

	############################################################################
################################################################################




# optionsbar_rotate.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from .abstract_optionsbar import AbstractOptionsBar
from .utilities_units import utilities_add_unit_to_spinbtn

class OptionsBarRotate(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarRotate'

	def __init__(self, rotate_tool):
		super().__init__()
		# knowing the tool is needed because the pane doesn't compact the same
		# way if it's applied to the selection
		self.rotate_tool = rotate_tool

		self._build_ui('transform/abstract-optionsbar-transform.ui')
		builder = self._hydrate_transform_tool('transform/optionsbar-rotate.ui')

		self.angle_btn = builder.get_object('angle_btn')
		utilities_add_unit_to_spinbtn(self.angle_btn, 3, '°')

		self.angle_box = builder.get_object('angle_box')
		self.rotate_box = builder.get_object('rotate_box')
		self.flip_box = builder.get_object('flip_box')

	def init_adaptability(self):
		super().init_adaptability()
		temp_limit_size = self.centered_box.get_preferred_width()[0] + \
		                    self.cancel_btn.get_preferred_width()[0] + \
		                      self.help_btn.get_preferred_width()[0] + \
		                     self.apply_btn.get_preferred_width()[0]
		self._set_limit_size(temp_limit_size)

	def update_for_new_tool(self, tool):
		self.set_compact(self._is_narrow)

	def set_compact(self, state):
		super().set_compact(state)
		if self.rotate_tool.apply_to_selection:
			self.options_btn.set_visible(state)
			self.angle_box.set_visible(True)
			self.rotate_box.set_visible(not state)
			self.flip_box.set_visible(not state)
		else:
			self.options_btn.set_visible(False)
			self.angle_box.set_visible(False)
			self.rotate_box.set_visible(True)
			self.flip_box.set_visible(True)

	############################################################################
################################################################################




# optionsbar_skew.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from .abstract_optionsbar import AbstractOptionsBar
from .utilities_units import utilities_add_unit_to_spinbtn

class OptionsBarSkew(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarSkew'

	def __init__(self):
		super().__init__()

		self._build_ui('transform/abstract-optionsbar-transform.ui')
		builder = self._hydrate_transform_tool('transform/optionsbar-skew.ui')

		self.xy_label = builder.get_object('xy_label')
		self.yx_label = builder.get_object('yx_label')
		self.separator = builder.get_object('separator')

		self.yx_spinbtn = builder.get_object('yx_spinbtn')
		self.xy_spinbtn = builder.get_object('xy_spinbtn')
		utilities_add_unit_to_spinbtn(self.yx_spinbtn, 3, '%')
		utilities_add_unit_to_spinbtn(self.xy_spinbtn, 3, '%')

	def init_adaptability(self):
		super().init_adaptability()
		temp_limit_size = self.centered_box.get_preferred_width()[0] + \
		                    self.cancel_btn.get_preferred_width()[0] + \
		                   self.options_btn.get_preferred_width()[0] + \
		                      self.help_btn.get_preferred_width()[0] + \
		                     self.apply_btn.get_preferred_width()[0]
		self._set_limit_size(temp_limit_size)

	def update_for_new_tool(self, tool):
		self.set_compact(self._is_narrow)

	def set_compact(self, state):
		super().set_compact(state)
		if state:
			self.centered_box.set_orientation(Gtk.Orientation.VERTICAL)
		else:
			self.centered_box.set_orientation(Gtk.Orientation.HORIZONTAL)
		self.xy_label.set_visible(not state)
		self.yx_label.set_visible(not state)
		self.separator.set_visible(not state)

	############################################################################
################################################################################




# tool_text.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk, Gdk, GLib, Pango, PangoCairo
from .abstract_classic_tool import AbstractClassicTool

class ToolText(AbstractClassicTool):
	__gtype_name__ = 'ToolText'

	def __init__(self, window, **kwargs):
		super().__init__('text', _("Text"), 'tool-text-symbolic', window)

		self._should_cancel = False
		self._last_click_btn = 1

		self.add_tool_action_simple('text-set-font', self._set_font)
		self.add_tool_action_boolean('text-bold', False)
		self.add_tool_action_boolean('text-italic', False)

		self._background_id = self.load_tool_action_enum('text-background', \
		                                                 'last-text-background')
		self._font_fam_name = self.load_tool_action_enum('text-active-family', \
		                                                 'last-font-name')

		# XXX actions sensitivity?
		self.add_tool_action_simple('text-cancel', self._on_cancel)
		self.add_tool_action_simple('text-preview', self._force_refresh)
		self.add_tool_action_simple('text-insert', self._on_insert_text)

		builder = Gtk.Builder().new_from_resource(self.UI_PATH + 'tool-text.ui')
		# Widgets for text insertion
		self._popover = builder.get_object('insertion-popover')
		self._entry = builder.get_object('entry')
		self._entry.set_size_request(100, 50)
		self._entry.get_buffer().connect('changed', self._preview_text)
		self._hide_entry()

	############################################################################
	# Options ##################################################################

	def _set_font(self, *args):
		dialog = Gtk.FontChooserDialog(show_preview_entry=False)
		dialog.set_level(Gtk.FontChooserLevel.FAMILY)
		dialog.set_font(self._font_fam_name)

		# for f in PangoCairo.font_map_get_default().list_families():
		# 	print(f.get_name())

		try:
			status = dialog.run()
			# ^ FIXME #479
		except Exception as ex:
			# probablement qu'on ne catchera rien
			pass
		if(status == Gtk.ResponseType.OK):
			self._font_fam_name = dialog.get_font_family().get_name()
			# print(dialog.get_font())
			font_gvar = GLib.Variant.new_string(self._font_fam_name)
			self.window.lookup_action('text-active-family').set_state(font_gvar)
			self._preview_text()
		dialog.destroy()
		# TODO update le editing tip !!

	def _set_font_options(self, *args):
		# XXX incomplete? OBLIQUE exists
		self._is_italic = self.get_option_value('text-italic')
		self._is_bold = self.get_option_value('text-bold')

	def _set_background_style(self, *args):
		self._background_id = self.get_option_value('text-background')

	def get_options_label(self):
		return _("Text options")

	def get_editing_tips(self):
		self._set_font_options()
		self._set_background_style()

		# get_editing_tips is likely called because an option changed
		self._force_refresh()

		label_options = self.label + " - " + self._font_fam_name
		if self._background_id != 'none':
			bg_label = {
				'none': _("No background"),
				'shadow': _("Shadow"),
				'thin-outline': _("Thin outline"),
				'thick-outline': _("Thick outline"),
				'rectangle': _("Rectangle background"),
			}[self._background_id]
			label_options += " - " + bg_label

		return [label_options]

	############################################################################

	def on_tool_selected(self):
		super().on_tool_selected()
		self._last_click_btn = 1

	def on_tool_unselected(self):
		self.set_action_sensitivity('paste', True)
		self.set_action_sensitivity('select_all', True)
		self.set_action_sensitivity('selection_cut', True)
		self.set_action_sensitivity('selection_copy', True)

	def give_back_control(self, preserve_selection):
		if self._should_cancel:
			self._on_cancel()

	def force_text_tool(self, string):
		self.select_flowbox_child()
		self.set_common_values(self._last_click_btn, 100, 100)
		self._open_popover_at(100, 100)
		self._set_string(string)

	def _set_string(self, string):
		if string is None:
			string = ''
		self._entry.get_buffer().set_text(string, -1)

	############################################################################

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._last_click_btn = event.button
		self._should_cancel = True
		self.set_common_values(self._last_click_btn, event_x, event_y)
		self._open_popover_at(event.x, event.y)

	# XXX could there be a better way to move the text ?
	def _open_popover_at(self, x, y):
		rectangle = Gdk.Rectangle()
		rectangle.x = x
		rectangle.y = y
		rectangle.height = 1
		rectangle.width = 1
		self._popover.set_pointing_to(rectangle)
		self._popover.set_relative_to(self.get_image())
		self._popover.popup()
		self._entry.grab_focus()
		self._preview_text()

		# Usual text entry shortcuts don't work otherwise
		self.set_action_sensitivity('paste', False)
		self.set_action_sensitivity('select_all', False)
		self.set_action_sensitivity('selection_cut', False)
		self.set_action_sensitivity('selection_copy', False)

	def _hide_entry(self):
		self._popover.popdown()

	def _force_refresh(self, *args):
		self.set_common_values(self._last_click_btn, self.x_press, self.y_press)
		self._preview_text()

	def _on_insert_text(self, *args):
		self._hide_entry()
		if self._has_current_text():
			operation = self.build_operation()
			self.apply_operation(operation)
			self._set_string('')
		self.set_action_sensitivity('paste', True)
		self.set_action_sensitivity('select_all', True)
		self.set_action_sensitivity('selection_cut', True)
		self.set_action_sensitivity('selection_copy', True)

	def _has_current_text(self):
		b = self._entry.get_buffer()
		self.text_string = b.get_text(b.get_start_iter(), b.get_end_iter(), False)
		if self.text_string == '':
			self.restore_pixbuf()
			self.non_destructive_show_modif()
			return False
		else:
			return True

	def _preview_text(self, *args):
		if self._has_current_text():
			operation = self.build_operation()
			self.do_tool_operation(operation)

	def _on_cancel(self, *args):
		self._hide_entry()
		self._set_string('')
		self._should_cancel = False

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'rgba1': self.main_color,
			'rgba2': self.secondary_color,
			'font_fam': self._font_fam_name,
			'is_italic': self._is_italic,
			'is_bold': self._is_bold,
			'font_size': self.tool_width,
			'antialias': self._use_antialias,
			'x': self.x_press,
			'y': self.y_press,
			'background': self._background_id,
			'text': self.text_string
		}
		return operation

	def do_tool_operation(self, operation):
		cairo_context = self.start_tool_operation(operation)

		font_fam = operation['font_fam']
		font_size = operation['font_size'] * 2
		entire_text = operation['text']
		background_color = operation['rgba2']
		text_x = operation['x']
		text_y = operation['y']

		font_description_string = font_fam
		if operation['is_italic']:
			font_description_string += " Italic"
		if operation['is_bold']:
			font_description_string += " Bold"
		font_description_string += " " + str(font_size)
		font = Pango.FontDescription(font_description_string)

		p_context = PangoCairo.create_context(cairo_context)
		layout = Pango.Layout(p_context)
		layout.set_font_description(font)

		if not operation['antialias']:
			font_options = cairo.FontOptions()
			font_options.set_antialias(cairo.Antialias.NONE)
			font_options.set_hint_metrics(cairo.HintMetrics.OFF)
			font_options.set_hint_style(cairo.HintStyle.FULL)
			PangoCairo.context_set_font_options(p_context, font_options)

		########################################################################
		# Draw background ######################################################

		if operation['background'] == 'rectangle':
			lines = entire_text.split('\n')
			line_y = text_y
			for line_text in lines:
				line_y = self._op_bg_rectangle(cairo_context, layout, \
				                    background_color, text_x, line_y, line_text)
		elif operation['background'] == 'shadow':
			dist = max(min(int(font_size/16), 4), 1)
			cairo_context.set_source_rgba(*background_color)
			self._show_text_at_coords(cairo_context, layout, entire_text, \
			                                       text_x + dist, text_y + dist)
		elif operation['background'] == 'thin-outline':
			cairo_context.set_source_rgba(*background_color)
			dist = min(int(font_size/16), 10)
			dist = max(dist, 2)
			for dx in range(-dist, dist):
				for dy in range(-dist, dist):
					if abs(dx) + abs(dy) <= dist * 1.5:
						self._show_text_at_coords(cairo_context, layout, \
						                  entire_text, text_x + dx, text_y + dy)
			# these `for`s and this `if` should outline with an octogonal shape,
			# which is close enough to a smooth round outline imho.
		elif operation['background'] == 'thick-outline':
			cairo_context.set_source_rgba(*background_color)
			dist = int(font_size/10)
			dist = max(dist, 2)
			for dx in range(-dist, dist):
				for dy in range(-dist, dist):
					if abs(dx) + abs(dy) <= dist * 1.5:
						self._show_text_at_coords(cairo_context, layout, \
						                  entire_text, text_x + dx, text_y + dy)
			# looks better, but so much computation for bullshit...

		########################################################################
		# Draw text ############################################################

		cairo_context.set_source_rgba(*operation['rgba1'])
		self._show_text_at_coords(cairo_context, layout, entire_text, \
		                                                         text_x, text_y)

		self.non_destructive_show_modif()

	def _show_text_at_coords(self, cairo_c, pango_l, text, text_x, text_y):
		"""Use a pango layout (pango_l) to show a line of text (text) on a cairo
		context (cairo_c) at given coordinates (text_x, text_y)."""
		cairo_c.move_to(text_x, text_y)
		pango_l.set_text(text, -1)
		PangoCairo.update_layout(cairo_c, pango_l)
		PangoCairo.show_layout(cairo_c, pango_l)

	def _op_bg_rectangle(self, context, layout, c2, text_x, line_y, line_text):
		# The text is first "displayed" in a transparent color…
		context.set_source_rgba(0.0, 0.0, 0.0, 0.0)
		self._show_text_at_coords(context, layout, line_text, text_x, line_y)
		# …so we can get the size of the displayed line…
		ink_rect, logical_rect = layout.get_pixel_extents()
		delta_y = logical_rect.height
		# …and draw the background rectangle with the correct size.
		context.rel_line_to(0, delta_y)
		context.rel_line_to(logical_rect.width, 0)
		context.rel_line_to(0, -1 * delta_y)
		context.close_path()
		context.set_source_rgba(*c2)
		context.fill()
		context.stroke()
		# The returned value will be used as the vertical coord of the next line
		return line_y + delta_y

	############################################################################
################################################################################




# saving_manager.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GdkPixbuf, GLib
from .message_dialog import DrMessageDialog
from .utilities_files import utilities_add_filechooser_filters
from .utilities_colors import utilities_rgb_to_hexadecimal

ALL_SUPPORTED_FORMAT = ['jpeg', 'jpg', 'jpe', 'png', 'tiff', 'ico', 'bmp']

################################################################################

class DrSavingManager():
	__gtype_name__ = 'DrSavingManager'

	def __init__(self, window):
		self._window = window

	def save_current_image(self, is_export, to_new, selection_only, allow_alpha):
		"""All parameters are booleans. The value of `allow_alpha` reflects the
		user's intention (action they activated), not a technical capability.
		It returns a boolean (true if success)."""
		if not selection_only and not self._confirm_despite_ongoing_operation():
			return False

		if is_export:
			to_new = True

		image = self._window.get_active_image()

		if image.get_file_path() is None or to_new or selection_only:
			gfile = self._file_chooser_save()
		else:
			gfile = image.gfile

		if gfile is None:
			return False
		file_path = gfile.get_path()
		file_format = self._get_format(file_path)

		if selection_only:
			pixbuf = image.selection.get_pixbuf()
		else:
			pixbuf = image.main_pixbuf

		# Ask the user what to do concerning formats with no alpha channel
		if not allow_alpha:
			# the user clicked on "save without alpha", i'll not suggest the
			# exact opposite in the dialog
			can_save_as = False
		else:
			# no need to suggest "save as png" if it's an export, because
			# the file chooser dialog allows any format
			can_save_as = not is_export

		supports_alpha = file_format in ['png']
		if not allow_alpha or not supports_alpha:
			try:
				replacement = self._window.gsettings.get_string('replace-alpha')
				if replacement == 'ask':
					replacement = self._ask_overwrite_alpha(supports_alpha, can_save_as)
				pixbuf = self._replace_alpha(pixbuf, replacement, image)
			except WantAnotherFormatException as e:
				# the user wants to save the file under an other format
				return self.save_current_image(False, True, False, True)
			except WantToCancelException as e:
				# the user clicks on "cancel"
				return True
			except Exception as e:
				print(e) # an actual error occurred!
				# Context: an error message
				self._window.reveal_message(_("Failed to replace transparency"))

		# The "reload?" message shouldn't be shown, i force the reload later
		image.set_monitoring(False)

		try:
			# Actually save the pixbuf to the given file path
			pixbuf.savev(file_path, file_format, [None], [])

			# Update the image and the window objects
			if not is_export:
				image.gfile = gfile
				image.connect_gfile_monitoring()
				image.remember_current_state()
				image.post_save()
				self._window.update_picture_title()
		except Exception as e:
			image.set_monitoring(False)
			print(e)
			# Context: an error message
			self._window.reveal_message(_("Failed to save %s") % file_path)
			return False

		# Reset the file monitoring flag
		image.set_monitoring(True)

		if not is_export:
			# Update the image and the window objects
			try:
				image.gfile = gfile
				image.connect_gfile_monitoring()
				image.post_save()
				image.reload_from_disk()
			except Exception as e:
				print(e)
				# Context: an error message
				self._window.reveal_message(_("Failed to reload %s") % file_path)
				return False

		# everything went fine, return true
		return True

	############################################################################
	# Confirmation and file-chooser dialogs ####################################

	def confirm_save_modifs(self):
		"""Return True if the image can be closed/overwritten (whether it's
		saved or not), or False otherwise (usually if the user clicked 'cancel',
		or if an error occurred)."""
		image = self._window.get_active_image()
		if image.is_saved():
			return True
		fn = image.get_file_path()
		if fn is None:
			unsaved_file_name = _("Untitled") + '.png'
			# Context: the sentence "There are unsaved modifications to %s."
			display_name = _("this picture")
		else:
			unsaved_file_name = fn.split('/')[-1]
			display_name = image.get_filename_for_display()
		dialog = DrMessageDialog(self._window)
		discard_id = dialog.set_action(_("Discard"), 'destructive-action')
		cancel_id = dialog.set_action(_("Cancel"), None)
		save_id = dialog.set_action(_("Save"), None, True)
		dialog.add_string( _("There are unsaved modifications to %s.") % display_name)
		self._window.minimap.update_content()
		image = Gtk.Image().new_from_pixbuf(self._window.minimap.mini_pixbuf)
		frame = Gtk.Frame(valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER)
		frame.add(image)
		dialog.add_widget(frame)
		result = dialog.run()
		dialog.destroy()
		if result == save_id:
			return self._window.action_save()
		elif result == discard_id:
			return True
		else: # cancel_id
			return False

	def _confirm_despite_ongoing_operation(self):
		"""Ask to the user whether or not they want to apply the operation of
		the current tool (curve, shape, transform, selection) before saving."""
		msg = None
		if self._window.get_selection_tool().selection_is_active():
			msg = _("A part of the image is selected: the pixels beneath " + \
			   "the selection will be saved with a color you might not expect!")
		elif self._window.active_tool().has_ongoing_operation():
			# Context: the user tries to save the image while previewing an
			# unapplied "transform" operation (scaling, cropping, whatever)
			msg = _("Modifications from the current tool haven't been applied.")
		if msg is None:
			return True
		dialog = DrMessageDialog(self._window)

		# If we know how to finish the ongoing operation, then let's go
		can_apply = self._window.lookup_action('apply_transform').get_enabled()
		if can_apply:
			can_unselect = self._window.active_tool().apply_to_selection
		else:
			can_unselect = self._window.lookup_action('unselect').get_enabled()
		# Otherwise the user should cancel, and apply manually

		only_2_items = not (can_apply or can_unselect)
		cancel_id = dialog.set_action(_("Cancel"), None, only_2_items)
		if only_2_items:
			save_style = 'destructive-action'
		else:
			save_style = None
		apply_id = None
		if can_apply:
			# To translators: this string should be quite short
			apply_id = dialog.set_action(_("Apply & save"), \
			                                           'suggested-action', True)
			# same label whether or not `can_unselect` too ^
		elif can_unselect:
			# To translators: this string should be quite short
			apply_id = dialog.set_action(_("Deselect & save"), \
			                                           'suggested-action', True)
		save_id = dialog.set_action(_("Save anyway"), save_style)

		dialog.add_string(msg)
		dialog.add_string(_("Do you want to save anyway?"))
		self._window.minimap.update_content()
		image = Gtk.Image().new_from_pixbuf(self._window.minimap.mini_pixbuf)
		frame = Gtk.Frame(valign=Gtk.Align.CENTER, halign=Gtk.Align.CENTER)
		frame.add(image)
		dialog.add_widget(frame)
		result = dialog.run()
		dialog.destroy()

		if result == save_id:
			return True
		elif result == cancel_id:
			return False
		elif result == apply_id:
			if can_apply:
				self._window.action_apply_transform()
			if can_unselect:
				self._window.action_unselect()
			return True
		else: # unknown id
			return False

	def _file_chooser_save(self):
		"""Opens an "save" file chooser dialog, and return a GioFile or None."""
		gfile = None
		file_chooser = Gtk.FileChooserNative.new(_("Save picture as…"),
		       self._window, Gtk.FileChooserAction.SAVE, _("Save"), _("Cancel"))
		utilities_add_filechooser_filters(file_chooser)

		images_dir = GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES)
		if images_dir != None: # no idea why it sometimes fails
			file_chooser.set_current_folder(images_dir)
		# Context: Untitled(.png) is the default name of a newly saved file
		default_file_name = str(_("Untitled") + '.png')
		file_chooser.set_current_name(default_file_name)

		response = file_chooser.run()
		if response == Gtk.ResponseType.ACCEPT:
			gfile = file_chooser.get_file()
		file_chooser.destroy()
		return gfile

	############################################################################
	# Pixbuf transparency ######################################################

	def _get_format(self, file_path):
		"""Build a short string which will be recognized as a file format by the
		GdkPixbuf.Pixbuf.savev method."""
		file_format = file_path.split('.')[-1]
		file_format = file_format.lower()
		if file_format in ['jpeg', 'jpg', 'jpe']:
			file_format = 'jpeg'
		elif file_format not in ALL_SUPPORTED_FORMAT:
			file_format = 'png'
		return file_format

	def _ask_overwrite_alpha(self, supports_alpha, can_save_as):
		"""Warn the user about the replacement of the alpha channel for JPG or
		BMP files, but it may quickly annoy users to see a dialog so it's an
		option. Can be used on PNG files (if 'supports_alpha' is true)."""
		dialog = DrMessageDialog(self._window)
		cancel_id = dialog.set_action(_("Cancel"), None)
		if can_save_as:
			save_as_id = dialog.set_action(_("Save as…"), None)
		else:
			save_as_id = None
		# Context: confirm replacing transparent pixels with the selected color
		replace_id = dialog.set_action(_("Replace"), None, True)

		if not supports_alpha:
			dialog.add_string(_("This file format doesn't support transparent colors."))
		if can_save_as:
			dialog.add_string(_("You can save the image as a PNG file, or " \
			                                      "replace transparency with:"))
		else:
			dialog.add_string(_("Replace transparency with:"))

		alpha_combobox = Gtk.ComboBoxText(halign=Gtk.Align.CENTER)
		# Context: this is the label that describes the initial color of an
		# image (as set in the preferences or in the "new image with custom
		# size" dialog). This is displayed in the context of editing or saving
		# the image, but not when creating it.
		alpha_combobox.append('initial', _("Default color"))
		alpha_combobox.append('white', _("White"))
		alpha_combobox.append('black', _("Black"))
		# Context: replace transparent pixels with [checkboard]. This is a
		# pattern of dark and light greys.
		alpha_combobox.append('checkboard', _("Checkboard"))
		if not supports_alpha:
			# Context: replace transparent pixels with [nothing]. In practice it
			# means the alpha channel is just removed.
			alpha_combobox.append('nothing', _("Nothing"))
		alpha_combobox.set_active_id('initial') # If we run the dialog, it often
		# means the active preference is 'ask', so there is no way we can set
		# the default value to something more pertinent.
		dialog.add_widget(alpha_combobox)

		result = dialog.run()
		repl = alpha_combobox.get_active_id()
		dialog.destroy()
		if result == cancel_id:
			raise WantToCancelException()
		if result == save_as_id:
			raise WantAnotherFormatException()
		return repl

	def _replace_alpha(self, pixbuf, replacement, image):
		if replacement == 'nothing':
			# wouldn't work with a PNG file
			return pixbuf
		width = pixbuf.get_width()
		height = pixbuf.get_height()
		if replacement == 'white':
			pcolor1 = utilities_rgb_to_hexadecimal(255, 255, 255)
			pcolor2 = utilities_rgb_to_hexadecimal(255, 255, 255)
		elif replacement == 'initial':
			initial_rgba = image.get_initial_rgba()
			r = int(initial_rgba.red * 255)
			g = int(initial_rgba.green * 255)
			b = int(initial_rgba.blue * 255)
			# the initial color has an alpha channel but it's not pertinent, and
			# not possible anyway.
			pcolor1 = utilities_rgb_to_hexadecimal(r, g, b)
			pcolor2 = utilities_rgb_to_hexadecimal(r, g, b)
		elif replacement == 'checkboard':
			pcolor1 = utilities_rgb_to_hexadecimal(85, 85, 85)
			pcolor2 = utilities_rgb_to_hexadecimal(170, 170, 170)
		else: # if replacement == 'black':
			pcolor1 = utilities_rgb_to_hexadecimal(0, 0, 0)
			pcolor2 = utilities_rgb_to_hexadecimal(0, 0, 0)
		return pixbuf.composite_color_simple(width, height,
		                   GdkPixbuf.InterpType.TILES, 255, 8, pcolor1, pcolor2)

	############################################################################
################################################################################

class WantToCancelException(Exception):
	def __init__(self):
		super().__init__(_("Cancel"))

class WantAnotherFormatException(Exception):
	def __init__(self):
		super().__init__(_("Replace"))

################################################################################




# abstract_select.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk
from .abstract_tool import AbstractAbstractTool
from .optionsbar_selection import OptionsBarSelection
from .utilities_colors import utilities_gdk_rgba_to_normalized_array
from .utilities_overlay import utilities_show_overlay_on_context
from .selection_manager import NoSelectionPixbufException

class AbstractSelectionTool(AbstractAbstractTool):
	__gtype_name__ = 'AbstractSelectionTool'

	def __init__(self, tool_id, label, icon_name, window, **kwargs):
		super().__init__(tool_id, label, icon_name, window)
		self.menu_id = 2
		self.accept_selection = True

		self._future_pixbuf = None
		self.x_press = 0
		self.y_press = 0
		self.local_dx = 0
		self.local_dy = 0
		self.operation_type = None # 'op-define'

		self.load_tool_action_enum('selection-color', 'last-delete-replace')
		self.add_tool_action_boolean('selection-extract', False)

	############################################################################
	# UI implementations #######################################################

	def get_options_label(self):
		return _("Selection")

	def get_editing_tips(self):
		if self.selection_is_active():
			label_tip = _("Drag the selection or right-click on the canvas")
		else:
			label_tip = _("Select an area or right-click on the canvas")
		return [label_tip]

	def get_options_model(self):
		builder = Gtk.Builder.new_from_resource(self.UI_PATH + 'selection.ui')
		return builder.get_object('options-menu')

	def try_build_pane(self):
		self.pane_id = 'selection'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		return OptionsBarSelection(self.window)

	############################################################################
	# Lifecycle implementations ################################################

	def give_back_control(self, preserve_selection):
		self.get_selection().hide_popovers()
		# TODO if the selection hasn't been used, nothing should be applied
		if not preserve_selection:
			self.unselect_and_apply()

	# def on_tool_selected(self, *args):
	# 	pass

	# def on_tool_unselected(self, *args):
	# 	pass

	def cancel_ongoing_operation(self):
		pass # XXX really ?

	def has_ongoing_operation(self):
		return False

	############################################################################
	############################################################################

	def _get_press_behavior(self, event):
		if event.button == 3:
			return 'menu'
		elif not self.selection_is_active():
			return 'define'
		elif self.get_selection().point_is_in_selection(self.x_press, self.y_press):
			return 'drag'
		else:
			return 'cancel'

	def press_define(self, event_x, event_y):
		pass # implemented by actual tools

	def motion_define(self, event_x, event_y, render):
		pass # implemented by actual tools

	def release_define(self, surface, event_x, event_y):
		pass # implemented by actual tools

	############################################################################
	# Signal callbacks implementations #########################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.x_press = event_x
		self.y_press = event_y
		self.behavior = self._get_press_behavior(event)
		# print('press', self.behavior)
		if self.behavior == 'drag':
			self.cursor_name = 'grabbing'
			self.window.set_cursor(True)
		elif self.behavior == 'define':
			self.press_define(event_x, event_y)
		elif self.behavior == 'cancel':
			self.unselect_and_apply()
			self.restore_pixbuf()
			self.non_destructive_show_modif()

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if self.behavior == 'define':
			self.motion_define(event_x, event_y, render)
		elif render and self.behavior == 'drag':
			self._preview_drag_to(event_x, event_y)

	def _preview_drag_to(self, event_x, event_y):
		self.local_dx = event_x - self.x_press
		self.local_dy = event_y - self.y_press
		self.non_destructive_show_modif()

	def on_unclicked_motion_on_area(self, event, surface):
		x, y = self.get_image().get_event_coords(event)
		if not self.selection_is_active():
			self.cursor_name = 'cross'
		elif self.get_selection().point_is_in_selection(x, y):
			self.cursor_name = 'grab'
		else:
			self.cursor_name = 'cross'
		self.window.set_cursor(True)

	def on_release_on_area(self, event, surface, event_x, event_y):
		if event.button == 3:
			if not self.get_selection().is_active:
				self.get_selection().set_coords(True, event_x, event_y)
				# TODO je ne devrais jamais appeler cette méthode ^ ici
			self.get_selection().set_popovers_position(event.x, event.y)
			self.get_selection().show_popover()
			return
		self.restore_pixbuf()
		if self.behavior == 'define':
			self.release_define(surface, event_x, event_y)
		elif self.behavior == 'drag':
			self._apply_drag_to(event_x, event_y)

	def on_draw_above(self, area, ccontext):
		if not self.selection_is_active():
			return
		ldx = self.local_dx
		ldy = self.local_dy
		self.get_selection().show_selection_on_surface(ccontext, True, ldx, ldy)
		dragged_path = self.get_selection().get_path_with_scroll(ldx, ldy)
		# ^ Method not really use elsewhere, could it be private?
		thickness = self.get_overlay_thickness()
		utilities_show_overlay_on_context(ccontext, dragged_path, thickness)

	############################################################################
	# Pre-loading the selection manager with non-essential data ################

	# In the first phase of the tool execution (pre-building operations), the
	# code should never call methods from `self.get_selection().*` aside of:
	# - the getters (safe by definition)
	# - the things to show/hide popovers (no actual effect on the image)
	# - the `set_future_*` methods, which are designed to preload data whose
	# values don't have any impact on the following operations, but which should
	# stored image-wide anyway (to avoid inconsistencies when switching to other
	# tabs and then switching back, for example).
	#
	# XXX c'est faux le préchargement des coordonnées ça a un réel impact

	def _pre_load_coords(self, x, y):
		self.get_selection().set_future_coords(x, y)

	def _pre_load_path(self, path, resync_coords=True):
		self.get_selection().set_future_path(path, resync_coords)

	def _build_rectangle_path(self, press_x, press_y, release_x, release_y):
		"""Build rectangle path and pre-load it in the selection manager. This
		is used in `self.select_all` (abstract, here), and in the "rectangle
		selection" implementation."""
		# XXX could be in the selection manager?
		x0 = min(press_x, release_x)
		y0 = min(press_y, release_y)
		x1 = max(press_x, release_x)
		y1 = max(press_y, release_y)
		w = x1 - x0
		h = y1 - y0
		if w <= 0 or h <= 0:
			return
		self._pre_load_coords(x0, y0)
		cairo_context = self.get_context()
		cairo_context.new_path()
		cairo_context.move_to(x0, y0)
		cairo_context.line_to(x1, y0)
		cairo_context.line_to(x1, y1)
		cairo_context.line_to(x0, y1)
		cairo_context.close_path()
		self._pre_load_path(cairo_context.copy_path(), False)

	############################################################################
	# Operations producing methods #############################################

	def _apply_drag_to(self, event_x, event_y):
		x = self.get_selection().selection_x
		y = self.get_selection().selection_y
		fx = x + event_x - self.x_press
		fy = y + event_y - self.y_press
		self._pre_load_coords(fx, fy)

		self.operation_type = 'op-drag'
		operation = self.build_operation()
		self.do_tool_operation(operation)
		self.operation_type = 'op-define'

	def select_all(self):
		total_w = self.get_main_pixbuf().get_width()
		total_h = self.get_main_pixbuf().get_height()
		self._build_rectangle_path(0, 0, total_w, total_h)
		self.operation_type = 'op-define'
		operation = self.build_operation()
		self.apply_operation(operation)
		self.get_selection().show_popover()

	def delete_selection(self):
		self.operation_type = 'op-delete'
		operation = self.build_operation()
		self.apply_operation(operation)
		self.operation_type = 'op-define'

	def import_selection(self, pixbuf):
		self.unselect_and_apply()
		self._future_pixbuf = pixbuf
		self.operation_type = 'op-import'

		sx = self.get_selection().selection_x
		sy = self.get_selection().selection_y
		# when it's 0 and 0, it's pretty hard for the user to understand that
		# there is an active selection, but changing the coords here sucks, it
		# would be a pretty bad design, breaking the rules i've decided
		# concerning interactions with the selection manager.
		self._pre_load_coords(sx, sy)

		operation = self.build_operation()
		self.apply_operation(operation)
		self.operation_type = 'op-define'

	def replace_canvas(self):
		self.operation_type = 'op-replace-canvas'
		# XXX it's copied so many times?! ugly but seems necessary
		self._future_pixbuf = self.get_selection_pixbuf().copy()
		operation = self.build_operation()
		self.apply_operation(operation)
		self.operation_type = 'op-define'

	def invert_selection(self):
		pass # TODO

	def unselect_and_apply(self):
		# Pre-loading the coords is NEEDED because we may "unselect_and_apply" a
		# selection which was defined by a selection manager in very different
		# state, but was consumed by an "undo".
		sx = self.get_selection().selection_x
		sy = self.get_selection().selection_y
		self._pre_load_coords(sx, sy)

		self.operation_type = 'op-apply'
		operation = self.build_operation()
		self.apply_operation(operation)
		self.operation_type = 'op-define'
		self.cursor_name = 'cross'
		self.window.set_cursor(True)

	### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

	def build_operation(self):
		"""Operation is built from the operation_type, the current tool's
		'future_pixbuf' attribute, and '_future_*' attributes from the
		selection manager that were previously set."""
		if self._future_pixbuf is None: # Usual case
			pixbuf = None
		else: # Case of importation only
			pixbuf = self._future_pixbuf.copy()
			self._future_pixbuf = None

		replacement_type = self.get_option_value('selection-color')
		# FIXME the result is questionable when there was alpha in the area...
		if replacement_type == 'initial':
			gdk_rgba = self.get_image().get_initial_rgba()
			color = utilities_gdk_rgba_to_normalized_array(gdk_rgba)
		elif replacement_type == 'secondary':
			gdk_rgba = self.window.options_manager.get_right_color()
			color = utilities_gdk_rgba_to_normalized_array(gdk_rgba)
		else: # == 'alpha':
			color = [1.0, 1.0, 1.0, 0.0]
		# XXX the replacement is performed when the selection is defined, which
		# isn't the most intuitive behaviour: if the user sees a wrong
		# replacement color and changes the value to fix his manipulation, it's
		# already too late

		operation = {
			'tool_id': self.id,
			'operation_type': self.operation_type,
			'initial_path': self.get_selection().get_future_path(),
			'replacement': color,
			'extract': self.get_option_value('selection-extract'),
			'pixbuf': pixbuf,
			'pixb_x': self.get_selection().get_future_coords()[0],
			'pixb_y': self.get_selection().get_future_coords()[1],
			'local_dx': self.local_dx,
			'local_dy': self.local_dy,
		}
		return operation

	############################################################################
	# Operations consuming methods #############################################

	def _op_import(self, op):
		if op['pixbuf'] is None:
			raise NoSelectionPixbufException()
			# XXX does it run cleanly after the "return" to the calling method?
			# (compared to a more normal return)
		self._pre_load_coords(op['pixb_x'], op['pixb_y'])
		self.get_selection().set_coords(False, op['pixb_x'], op['pixb_y'])
		self.get_selection().set_pixbuf(op['pixbuf'].copy())

	def _op_replace_canvas(self, op):
		if op['pixbuf'] is None:
			raise NoSelectionPixbufException()
		self.get_image().set_main_pixbuf(op['pixbuf'].copy())
		self.get_image().use_stable_pixbuf()
		self.get_selection().reset(True)
		self.get_selection().reset_future_data()

	def _op_clean(self, operation):
		if operation['initial_path'] is None:
			return # The user double-clicked: there is no path, and it's normal
		cairo_context = self.get_context()
		cairo_context.new_path()
		cairo_context.append_path(operation['initial_path'])
		replacement_rgba = operation['replacement']
		cairo_context.set_operator(cairo.Operator.SOURCE)
		cairo_context.set_source_rgba(*replacement_rgba)
		cairo_context.fill()
		cairo_context.set_operator(cairo.Operator.OVER)

	def _op_drag(self, op):
		# print("drag to :", op['pixb_x'], op['pixb_y'])
		self.get_selection().set_coords(False, op['pixb_x'], op['pixb_y'])
		self.non_destructive_show_modif()

	def _op_define(self, op):
		if op['initial_path'] is None:
			return # The user double-clicked: there is no path, and it's normal
		self.get_selection().set_coords(True, op['pixb_x'], op['pixb_y'])
		if op['extract']:
			replacement = op['replacement']
		else:
			replacement = None
		self.get_selection().load_from_path(op['initial_path'], replacement)

	def _op_apply(self, operation):
		cairo_context = self.get_context()
		self.get_selection().show_selection_on_surface(cairo_context, False, \
		                           operation['local_dx'], operation['local_dy'])
		self.get_selection().reset(True)
		self.get_selection().reset_future_data()

	### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		# print('operation_type', operation['operation_type'])
		if operation['operation_type'] == 'op-delete':
			# Opération instantanée (sans preview), correspondant à une action
			# de type "clic-droit > couper" ou "clic-droit > supprimer".
			# On réinitialise le selection_manager.
			self.get_selection().reset(True)
		elif operation['operation_type'] == 'op-import':
			# Opération instantanée (sans preview), correspondant à une action
			# de type "clic-droit > importer" ou "clic-droit > coller".
			# On charge un pixbuf dans le selection_manager.
			self._op_import(operation)
		elif operation['operation_type'] == 'op-replace-canvas':
			# TODO commentaire
			self._op_replace_canvas(operation)
		elif operation['operation_type'] == 'op-define':
			# Opération instantanée (sans preview), correspondant à une
			# sélection (rectangulaire ou non) par définition d'un path.
			# On charge un pixbuf dans le selection_manager.
			self._op_define(operation)
			self._op_clean(operation)
		elif operation['operation_type'] == 'op-drag':
			# Prévisualisation d'opération, correspondant à la définition d'une
			# sélection (rectangulaire ou non) par construction d'un path.
			# On modifie les coordonnées connues du selection_manager.
			self.local_dx = 0
			self.local_dy = 0
			self._op_drag(operation)
		elif operation['operation_type'] == 'op-apply':
			# Opération instantanée correspondant à l'aperçu de l'op-drag, donc
			# la définition d'une sélection (rectangulaire ou non) par
			# construction d'un path qui sera "fusionné" au main_pixbuf.
			# On modifie les coordonnées connues du selection_manager.
			if self.get_selection_pixbuf() is None:
				return
			self._op_drag(operation)
			self._op_apply(operation)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from gi.repository import Gdk
from .abstract_filter import AbstractFilter

class FilterVeil(AbstractFilter):
	__gtype_name__ = 'FilterVeil'

	def do_filter_operation(self, source_pixbuf, operation):
		self._tool.get_image().set_temp_pixbuf(source_pixbuf.copy())
		temp = self._tool.get_image().temp_pixbuf
		source_pixbuf.saturate_and_pixelate(temp, 1, True)

	############################################################################
################################################################################




# selection_manager.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk, Gdk, GdkPixbuf

class NoSelectionPixbufException(Exception):
	def __init__(self, *args):
		# Context: an error message
		super().__init__(_("The selection pixbuf is empty."))

class NoSelectionPathException(Exception):
	def __init__(self, *args):
		super().__init__(_("The selection path is empty."))

################################################################################

class DrSelectionManager():
	__gtype_name__ = 'DrSelectionManager'

	def __init__(self, image):
		self.image = image
		self.init_pixbuf()
		self.reset_future_data()

		builder = Gtk.Builder.new_from_resource( \
		                '/com/github/maoschanz/drawing/ui/selection-manager.ui')
		menu_r = builder.get_object('inactive-selection-menu')
		self.menu_if_inactive = Gtk.Popover.new_from_model(self.image, menu_r)
		menu_l = builder.get_object('active-selection-menu')
		self.menu_if_active = Gtk.Popover.new_from_model(self.image, menu_l)
		self.set_popovers_position(1.0, 1.0)

	def init_pixbuf(self):
		# print('⇒ init pixbuf')
		self.selection_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, \
		                                                          True, 8, 1, 1)
		self.set_coords(True, 0, 0)
		self.selection_path = None
		self.is_active = False

	def load_from_path(self, new_path, rgba=None):
		"""Create a selection_pixbuf from a minimal part of the main surface by
		erasing everything outside of the provided path."""
		if new_path is None:
			raise NoSelectionPathException()

		self.selection_path = new_path
		self.is_active = True
		main_pixbuf = self.image.main_pixbuf

		# Erase everything outside of the path
		surface = Gdk.cairo_surface_create_from_pixbuf(main_pixbuf, 0, None)
		surface.set_device_scale(self.image.SCALE_FACTOR, self.image.SCALE_FACTOR)
		cairo_context = cairo.Context(surface)
		cairo_context.set_operator(cairo.Operator.DEST_IN)
		cairo_context.new_path()
		cairo_context.append_path(self.selection_path)
		cairo_context.fill_preserve()
		cairo_context.set_operator(cairo.Operator.OVER)

		# Find the coords to reduce the size of what will be stored
		main_width = main_pixbuf.get_width()
		main_height = main_pixbuf.get_height()
		xmin, ymin, xmax, ymax = cairo_context.path_extents()
		xmax = min(xmax, main_width)
		ymax = min(ymax, main_height)
		xmin = int( max(xmin, 0.0) ) # If everything is right, this is selection_x
		ymin = int( max(ymin, 0.0) ) # If everything is right, this is selection_y
		if self.selection_x < 0:
			xmin = self.selection_x
		if self.selection_y < 0:
			ymin = self.selection_y

		# Actually store the pixbuf
		selection_width = int(xmax - xmin)
		selection_height = int(ymax - ymin)
		if selection_width > 0 and selection_height > 0:
			# print('⇒ load pixbuf')
			pixbuf = Gdk.pixbuf_get_from_surface(surface, int(xmin), int(ymin),\
			                                  selection_width, selection_height)
			if pixbuf is not None:
				if rgba is not None and rgba[3] > 0.0:
					pixbuf = pixbuf.add_alpha(True, int(rgba[0] * 255), \
					                                int(rgba[1] * 255), \
					                                int(rgba[2] * 255))
				self.selection_pixbuf = pixbuf
			# can't use `set_pixbuf` here ^ because it would replace the free
			# path with a rectangle path
		else:
			self.reset(True)
		self.image.update_actions_state()

	def set_coords(self, temp_too, x, y):
		self.selection_x = x
		self.selection_y = y
		if temp_too:
			self.temp_x = x
			self.temp_y = y

	def set_pixbuf(self, pixbuf):
		# print('⇒ set pixbuf')
		self.selection_pixbuf = pixbuf
		self._create_path_from_pixbuf()

	def get_pixbuf(self):
		return self.selection_pixbuf

	def reset(self, update_image):
		# print('⇒ reset pixbuf')
		self.selection_pixbuf = None
		self.selection_path = None
		self.set_coords(True, 0, 0)
		self.is_active = False
		if update_image:
			self.image.update_actions_state()
			self.image.update()

	def get_path_with_scroll(self, tool_dx, tool_dy):
		# The very concept of this method sucks
		if self.selection_path is None:
			raise NoSelectionPathException()
		delta_x = tool_dx - self.image.scroll_x + self.selection_x - self.temp_x # XXX SHOULDN'T USE TEMP
		delta_y = tool_dy - self.image.scroll_y + self.selection_y - self.temp_y # XXX SHOULDN'T USE TEMP
		cairo_context = self._get_context_with_path(delta_x, delta_y)
		cairo_context.close_path()
		return cairo_context.copy_path()

	def show_selection_on_surface(self, cairo_context, with_scroll, tool_dx, tool_dy):
		if self.selection_pixbuf is None:
			raise NoSelectionPixbufException()
		if with_scroll:
			x = self.selection_x - self.image.scroll_x + tool_dx
			y = self.selection_y - self.image.scroll_y + tool_dy
		else:
			x = self.selection_x + tool_dx
			y = self.selection_y + tool_dy
		Gdk.cairo_set_source_pixbuf(cairo_context, self.selection_pixbuf, x, y)
		if self.image.is_zoomed_surface_sharp():
			cairo_context.get_source().set_filter(cairo.FILTER_NEAREST)
		cairo_context.paint()

	def get_center_coords(self):
		"""Return the coords of the center of the selection."""
		w = self.selection_pixbuf.get_width()
		h = self.selection_pixbuf.get_height()
		return self.selection_x + w / 2, self.selection_y + h / 2

	def point_is_in_selection(self, tested_x, tested_y):
		"""Returns a boolean if the point whose coordinates are "(tested_x,
		tested_y)" is in the path defining the selection. If such path doesn't
		exist, it returns None."""
		if not self.is_active:
			return True # shouldn't happen
		scrolled_path = self.get_path_with_scroll(self.image.scroll_x, self.image.scroll_y)
		cairo_context = self._get_context()
		cairo_context.new_path()
		cairo_context.append_path(scrolled_path)
		return cairo_context.in_fill(tested_x, tested_y)

	############################################################################

	def _create_path_from_pixbuf(self):
		"""This method creates a rectangle selection from the currently set
		selection pixbuf.
		It can be the result of an editing operation (crop, scale, etc.), or it
		can be an imported picture (from a file or from the clipboard)."""
		if self.selection_pixbuf is None:
			raise NoSelectionPixbufException()
		self.temp_x = self.selection_x
		self.temp_y = self.selection_y
		self.is_active = True
		cairo_context = self._get_context()
		cairo_context.move_to(self.selection_x, self.selection_y)
		cairo_context.rel_line_to(self.selection_pixbuf.get_width(), 0)
		cairo_context.rel_line_to(0, self.selection_pixbuf.get_height())
		cairo_context.rel_line_to(-1 * self.selection_pixbuf.get_width(), 0)
		cairo_context.close_path()
		self.selection_path = cairo_context.copy_path()
		self.hide_popovers()
		self.image.update_actions_state()

	def _get_context(self):
		return cairo.Context(self.image.surface)

	def _get_context_with_path(self, delta_x, delta_y):
		cairo_context = self._get_context()
		for pts in self.selection_path:
			if pts[1] != ():
				x = pts[1][0] + delta_x
				y = pts[1][1] + delta_y
				cairo_context.line_to(int(x), int(y))
		return cairo_context

	############################################################################
	# Popover menus management methods #########################################

	def set_popovers_position(self, x, y):
		"""Set the coords where the popover should be opened. These coords are
		relative TO THE WIDGET, not to the pixbuf."""
		rectangle = Gdk.Rectangle()
		rectangle.x = int(x)
		rectangle.y = int(y)
		rectangle.height = 1
		rectangle.width = 1
		self.menu_if_inactive.set_pointing_to(rectangle)
		self.menu_if_active.set_pointing_to(rectangle)

	def hide_popovers(self):
		self.menu_if_active.popdown()
		self.menu_if_inactive.popdown()

	def show_popover(self):
		"""Open the adequate popover at the previously set coords."""
		if self.is_active:
			self.menu_if_active.popup()
		else:
			self.menu_if_inactive.popup()

	############################################################################
	# Future path/coords #######################################################

	def reset_future_data(self):
		self._future_x = 0
		self._future_y = 0
		self._future_path = None

	def set_future_coords(self, x, y):
		self._future_x = int(x)
		self._future_y = int(y)

	def get_future_coords(self):
		return self._future_x, self._future_y

	def set_future_path(self, path, resync_coords):
		self._future_path = path

		if not resync_coords:
			return
		# Convert selection manager's future_path coords from absolute to
		# relative ones, and sets future coords accordingly.
		main_width = self.image.main_pixbuf.get_width()
		main_height = self.image.main_pixbuf.get_height()
		xmin, ymin = main_width, main_height # TODO context.path_extents() ?
		for pts in self._future_path:
			if pts[1] != ():
				xmin = min(pts[1][0], xmin)
				ymin = min(pts[1][1], ymin)
		self.set_future_coords(max(xmin, 0.0), max(ymin, 0.0))

	def get_future_path(self):
		return self._future_path

	def update_from_transform_tool(self, new_pixbuf, dx, dy):
		self.set_pixbuf(new_pixbuf)
		x = self.selection_x + dx
		y = self.selection_y + dy
		self.set_coords(False, x, y)
		self.set_future_coords(self._future_x + dx, self._future_y + dy)

	############################################################################
	# Debug ####################################################################

	def print_values(self, *args):
		"""Development only: is linked to the "troubleshoot selection" item in
		the short hamburger menu."""
		print("selection coords", self.selection_x, self.selection_y)
		print("temp coords", self.temp_x, self.temp_y)
		print("image.scroll coords", self.image.scroll_x, self.image.scroll_y)
		print("image.zoom_level", self.image.zoom_level)

		print("selection_path with scroll & temp deltas")
		delta_x = 0 - self.image.scroll_x + self.selection_x - self.temp_x
		delta_y = 0 - self.image.scroll_y + self.selection_y - self.temp_y
		for pts in self.selection_path:
			if pts[1] != ():
				x = pts[1][0] + delta_x
				y = pts[1][1] + delta_y
				print('\t', x, y)

		print("future path", self._future_path)
		print("future coords", self._future_x, self._future_y)
		print("---------------------------------------------------------------")

	############################################################################
################################################################################

